import os
import glob
import datetime as dt
import numpy as np
import pandas as pd
import xarray as xr
import xugrid as xu
from scipy.spatial import KDTree
import logging
import hydrolib.core.dflowfm as hcdfm
import geopandas

from dfm_tools.hydrolib_helpers import (Dataset_to_TimeSeries, 
                                        Dataset_to_T3D, 
                                        Dataset_to_Astronomic, 
                                        PolyFile_to_geodataframe_points, 
                                        get_ncbnd_construct,
                                        da_from_gdf_points,
                                        maybe_convert_fews_to_dfmt,
                                        validate_polyline_names)
from dfm_tools.errors import OutOfRangeError
from dfm_tools import settings

__all__ = ["get_conversion_dict",
           "interpolate_tide_to_bc",
           "interpolate_tide_to_plipoints",
           "interp_regularnc_to_plipointsDataset",
           "interp_uds_to_plipoints",
           "interp_hisnc_to_plipoints",
           "plipointsDataset_to_ForcingModel",
    ]

logger = logging.getLogger(__name__)


def get_conversion_dict(ncvarname_updates={}):
    
    """
    get the conversion_dict, optionally with updated ncvarnames
    conversion_dict.keys() are the dflowfm quantities and the ncvarname the corresponding netcdf variable name/key
    alternative ncvarnames can be supplied via ncvarname_updates, e.g. get_conversion_dict(ncvarname_updates={'temperaturebnd':'tos'})
    
    interpolate_nc_to_bc() renames netcdf variable like this:
    data_xr = data_xr.rename({ncvarname:quantity})
    
    for CMCC:
    conversion_dict = {
    # conversion is phyc in mol/m3 to newvar in g/m3
    'tracerbndOXY'        : {'ncvarname': 'o2',          'unit': 'g/m3', 'conversion': 32.0 },
    'tracerbndNO3'        : {'ncvarname': 'no3',         'unit': 'g/m3', 'conversion': 14.0 },
    'tracerbndPO4'        : {'ncvarname': 'po4',         'unit': 'g/m3', 'conversion': 30.97 },
    'tracerbndSi'         : {'ncvarname': 'si',          'unit': 'g/m3', 'conversion': 28.08},
    'tracerbndPON1'       : {'ncvarname': 'phyc',        'unit': 'g/m3', 'conversion': 14.0}, # Caution: this empirical relation might not be applicable to your use case
    'tracerbndPOP1'       : {'ncvarname': 'phyc',        'unit': 'g/m3', 'conversion': 30.97}, # Caution: this empirical relation might not be applicable to your use case
    'tracerbndPOC1'       : {'ncvarname': 'phyc',        'unit': 'g/m3', 'conversion': 14.0 * (106 / 16)}, # Caution: this empirical relation might not be applicable to your use case
    'salinitybnd'         : {'ncvarname': 'sos'},         #'1e-3'
    'temperaturebnd'      : {'ncvarname': 'tos'},         #'degC'
    'ux'                  : {'ncvarname': 'uo'},          #'m/s'
    'uy'                  : {'ncvarname': 'vo'},          #'m/s'
    'waterlevelbnd'       : {'ncvarname': 'zos'},         #'m' #steric
    'tide'                : {'ncvarname': ''},            #'m' #tide (dummy entry)
    }
    
    
    The below clarification about the CMEMS conversion factors in this function is by Jos van Gils.
    The use of boundary conditions from CMEMS requires some conversion (scaling). 
    In the first place, there is the unit conversion from mmol/m3 in CMEMS to g/m3 in DFM. 
    This involves a factor M/1000, where M is the molar weight of the constituent in question: 
    M = 12 for constituents expressed in gC/m3, 14 for gN/m3, 30.97 for gP/m3, 28.08 for gSi/m3 and 32 for dissolved oxygen. 
    
    Next there is the issue that CMEMS does not provide all state variables in the water 
    quality model. Dead organic matter, particulate and dissolved, are missing. 
    In practice, we use phytoplankton biomass from CMEMS (phyc) to estimate these missing variables. 
    For particulate organic matter (POM), often a carbon-based ratio of 2.0 to phytoplankton biomass is used.
    PON and POP are derived from POC by classical Redfield ratios (C:N:P = 106:16:1) or revised Redfield ratios (C:N:P = 117:14:1). 
    For Opal (biogenic silica from diatoms), the example here is based on a factor 0.5 expressing the share of 
    diatoms in phytoplankton and a 0.13 Si:C stoichiometric ratio by Brzezinski (1985).
    For dissolved organic matter (DOM) there is limited experience. This
    conversion_dict uses the approach used for the “Bays Eutrophication Model” (Massachusetts Bay, 
    https://www.mwra.com/harbor/enquad/pdf/2021-02.pdf). This relies on local field data.
    
    It is noted that such scale factors are anyhow inaccurate and typically water-system-specific. 
    Therefore, the values included in this conversion_dict should be considered indicative only and 
    should by no means be seen as a default approach. Ideally, the scale factors to link missing 
    states to available CMEMS variables are derived from local field data. 
    A validation is strongly recommended, for example by a comparison of simulated and 
    measured concentrations of total N and total P in the study area. 
    Similar considerations would hold for other data sources than CMEMS.
    """
    
    # conversion_dict for CMEMS
    conversion_dict = { # conversion is phyc in mmol/m3 to newvar in g/m3
                        'tracerbndOXY'        : {'ncvarname': 'o2',          'unit': 'g/m3', 'conversion': 32/1000},
                        'tracerbndNO3'        : {'ncvarname': 'no3',         'unit': 'g/m3', 'conversion': 14/1000},
                        'tracerbndPO4'        : {'ncvarname': 'po4',         'unit': 'g/m3', 'conversion': 30.97/1000},
                        'tracerbndSi'         : {'ncvarname': 'si',          'unit': 'g/m3', 'conversion': 28.08/1000},
                        'tracerbndGreen'      : {'ncvarname': 'phyc',        'unit': 'g/m3', 'conversion': 12/1000},
                        'tracerbndPON1'       : {'ncvarname': 'phyc',        'unit': 'g/m3', 'conversion': 2 * 16/106 * 14/1000}, # Caution: this empirical relation might not be applicable to your use case
                        'tracerbndPOP1'       : {'ncvarname': 'phyc',        'unit': 'g/m3', 'conversion': 2 * 30.97 / (106 * 1000)}, # Caution: this empirical relation might not be applicable to your use case
                        'tracerbndPOC1'       : {'ncvarname': 'phyc',        'unit': 'g/m3', 'conversion': 2 * 12/1000}, # Caution: this empirical relation might not be applicable to your use case
                        'tracerbndDON'        : {'ncvarname': 'phyc',        'unit': 'g/m3', 'conversion': 3.24 * 2 * 16/106 * 14/1000}, # Caution: this empirical relation might not be applicable to your use case
                        'tracerbndDOP'        : {'ncvarname': 'phyc',        'unit': 'g/m3', 'conversion': 2 * 30.97 / (106 * 1000)}, # Caution: this empirical relation might not be applicable to your use case
                        'tracerbndDOC'        : {'ncvarname': 'phyc',        'unit': 'g/m3', 'conversion': (199 / 20) * 3.24 * 2 * 16 * 12 / (106 * 1000)}, # Caution: this empirical relation might not be applicable to your use case
                        'tracerbndOpal'       : {'ncvarname': 'phyc',        'unit': 'g/m3', 'conversion': 0.5 * 0.13 * 28.08/1000}, # Caution: this empirical relation might not be applicable to your use case
                        'salinitybnd'         : {'ncvarname': 'so'},          #'1e-3'
                        'temperaturebnd'      : {'ncvarname': 'thetao'},      #'degC'
                        'ux'                  : {'ncvarname': 'uo'},          #'m/s'
                        'uy'                  : {'ncvarname': 'vo'},          #'m/s'
                        'waterlevelbnd'       : {'ncvarname': 'zos'},         #'m' #steric
                        'tide'                : {'ncvarname': ''},            #'m' #tide (dummy entry)
                        }
    #do updates
    for k,v in ncvarname_updates.items():
        conversion_dict[k]['ncvarname'] = v
    return conversion_dict


def ext_add_boundary_object_per_polyline(ext_new:hcdfm.ExtModel, boundary_object:hcdfm.Boundary):
    """
    If the polyfile contains multiple polylines, only the first 
    one will be used by DFLOW-FM for the boundary conditions.
    This function will save each polyline as a separate polyfile 
    and add duplicate boundary entries for each polyline to the extfile.
    """
    file_pli_main = boundary_object.locationfile.filepath
    polyfile_obj = hcdfm.PolyFile(file_pli_main)
    validate_polyline_names(polyfile_obj)
    dir_output = os.path.dirname(file_pli_main)
    for polyline_obj in polyfile_obj.objects:
        if len(polyfile_obj.objects) > 1:
            # copy to avoid location file to be the same for all duplicated boundary objects
            boundary_object = boundary_object.copy()
            # create a polyfile with a single plifile
            polyfile_oneline_obj = hcdfm.PolyFile(objects=[polyline_obj])
            file_pli_oneline = os.path.join(dir_output, f"{polyline_obj.metadata.name}.pli")
            if str(file_pli_oneline) == str(file_pli_main):
                raise ValueError("The names of one of the polylines in the polyfile is the "
                                 "same as the polyfilename, resulting in the same filename")
            polyfile_oneline_obj.save(file_pli_oneline)
            # update locationfile to the plifile with a single polyline
            boundary_object.locationfile = file_pli_oneline
        # add boundary to extmodel
        ext_new.boundary.append(boundary_object)


def interpolate_tide_to_bc(ext_new: hcdfm.ExtModel, tidemodel, file_pli, component_list=None, load=True):
    # read polyfile as geodataframe
    polyfile_obj = hcdfm.PolyFile(file_pli)
    gdf_points = PolyFile_to_geodataframe_points(polyfile_obj)
    
    # interpolate tidal components to plipoints
    data_interp = interpolate_tide_to_plipoints(tidemodel=tidemodel, gdf_points=gdf_points, 
                                                component_list=component_list, load=load)
    
    # convert interpolated xarray.Dataset to hydrolib ForcingModel and save as bc file
    ForcingModel_object = plipointsDataset_to_ForcingModel(plipointsDataset=data_interp)
    dir_output = os.path.dirname(file_pli)
    file_pli_name = polyfile_obj.filepath.stem
    file_bc_out = os.path.join(dir_output,f'tide_{tidemodel}_{file_pli_name}.bc')
    ForcingModel_object.save(filepath=file_bc_out)
    
    # generate hydrolib-core Boundary object to be appended to the ext file
    boundary_object = hcdfm.Boundary(quantity='waterlevelbnd', #the FM quantity for tide is also waterlevelbnd
                                      locationfile=file_pli,
                                      forcingfile=ForcingModel_object)
    
    # add the boundary object to the ext file for each polyline in the polyfile
    ext_add_boundary_object_per_polyline(ext_new=ext_new, boundary_object=boundary_object)


def tidemodel_componentlist(tidemodel:str, convention:bool):
    
    comp_list_dict = {'FES2014':['2n2','eps2','j1','k1','k2','l2','la2','m2','m3','m4','m6','m8','mf','mks2','mm','mn4','ms4','msf','msqm','mtm','mu2','n2','n4','nu2','o1','p1','q1','r2','s1','s2','s4','sa','ssa','t2'],
                      'FES2012':['2N2','E2',  'J1','K1','K2','L2','La2','M2','M3','M4','M6','M8','Mf','MKS2','Mm','MN4','MS4','MSf',       'Mtm','Mu2','N2','N4','Nu2','O1','P1','Q1','R2','S1','S2','S4',     'Ssa','T2','Z0'],
                      'EOT20':['2N2','J1','K1','K2','M2','M4','MF','MM','N2','O1','P1','Q1','S1','S2','SA','SSA','T2'],
                      'GTSMv4.1':['2N2','2Q1','CHI1','EPS2','ETA2','J1','K1','K2','L2','LABDA2','M1','M2','M3','M4','MF','MM','MSF','MSQM','MU2','N2','NU2','O1','OO1','P1','PI1','Q1','R2','S1','S2','SA','SSA','T2'],
                      'tpxo80_opendap':['M2','S2','N2','K2','K1','O1','P1','Q1','MF','MM','M4','MS4','MN4'],
                      }
    comp_list_dict['GTSMv4.1_opendap'] = comp_list_dict['GTSMv4.1']
    
    component_list = comp_list_dict[tidemodel]
    
    if convention:
        return components_translate_upper(component_list)
    else:
        return component_list


def components_translate_upper(component_list):
    # translate dict from .\hydro_tools\FES\PreProcessing_FES_TideModel_imaginary.m
    translate_dict = {'LA2':'LABDA2', #TODO: use value instead of key in bc file? Support using value instead of key in const_list also (like line above)
                      'EPS2':'EPSILON2', 
                      'Z0':'A0',
                      'MTM':'MFM', #Needs to be verified
                      #added later
                      'E2':'EPSILON2',
                      }
    
    component_list = pd.Series([x.upper() for x in component_list]).replace(translate_dict, regex=True).to_list()
    return component_list


def interpolate_tide_to_plipoints(tidemodel, gdf_points, component_list=None, load=True):
    """
    empty docstring
    """
    
    dir_pattern_dict = {
        'FES2014': settings.PATH_FES2014,
        'FES2012': settings.PATH_FES2012,
        'EOT20': settings.PATH_EOT20,
        'GTSMv4.1': settings.PATH_GTSMv41,
        'GTSMv4.1_opendap': settings.PATH_GTSMv41_opendap,
        'tpxo80_opendap': settings.PATH_tpxo80_opendap,
        }
    
    if tidemodel not in dir_pattern_dict.keys():
        raise KeyError(f"Invalid tidemodel '{tidemodel}', options are: {str(list(dir_pattern_dict.keys()))}")
        
    dir_pattern = dir_pattern_dict[tidemodel]

    component_list_tidemodel = tidemodel_componentlist(tidemodel, convention=False)
    component_list_tidemodel_convention = tidemodel_componentlist(tidemodel, convention=True)
    component_tidemodel_translate_dict = {k:v for k,v in zip(component_list_tidemodel_convention,component_list_tidemodel)}
    if component_list is None:
        component_list = component_list_tidemodel_convention
    else:
        component_list = components_translate_upper(component_list)
        
    def extract_component(ds):
        if tidemodel in ['FES2012']:
            ds = ds.sel(lon=ds.lon<360) #drop last instance, since 0 and 360 are both present
            ds = ds.rename({'Ha':'amplitude','Hg':'phase'})
            ds['phase'] = ds['phase'].assign_attrs({'units':'degrees'}) # degree in original file
        elif tidemodel in ['EOT20']:
            ds = ds.rename({'imag':'wl_imag','real':'wl_real'})
            ds = ds.sel(lon=ds.lon<360) #drop last instance, since 0 and 360 are both present
        elif tidemodel in ['GTSMv4.1','GTSMv4.1_opendap']:
            ds = ds.rename({'wl_amp':'amplitude','wl_phs':'phase'})
        
        convert_360to180 = (ds['lon'].to_numpy()>180).any()
        if convert_360to180: # results in large chunks if it is done after concatenation, so do for each file before concatenation
            ds.coords['lon'] = (ds.coords['lon'] + 180) % 360 - 180
            ds = ds.sortby('lon')
        return ds
    
    if tidemodel=='tpxo80_opendap':
        ds = xr.open_dataset(dir_pattern)
        bool_land = ds.depth==0
        ds['tidal_amplitude_h'] = ds['tidal_amplitude_h'].where(~bool_land).assign_attrs({'units':'m'})
        ds['tidal_phase_h'] = ds['tidal_phase_h'].where(~bool_land).assign_attrs({'units':'degrees'})
        convert_360to180 = (ds['lon'].to_numpy()>180).any()
        if convert_360to180:
            ds.coords['lon'] = (ds.coords['lon'] + 180) % 360 - 180
            ds = ds.sortby('lon')
        ds = ds.rename({'tidal_amplitude_h':'amplitude','tidal_phase_h':'phase','constituents':'compno'})
        ds = ds.drop_dims(['lon_u','lon_v','lat_u','lat_v'])
        components_infile = ds['tidal_constituents'].load().str.decode('utf-8',errors='ignore').str.strip().str.upper()
        ds['compnames'] = components_infile
        ds = ds.set_index({'compno':'compnames'})
        if component_list is not None:
            ds = ds.sel(compno=component_list)
        data_xrsel = ds
    else:
        #use open_mfdataset() with preprocess argument to open all requested tide files into one Dataset
        file_list_nc = [str(dir_pattern).replace('*',component_tidemodel_translate_dict[comp]) for comp in component_list]
        data_xrsel = xr.open_mfdataset(file_list_nc, combine='nested', concat_dim='compno', preprocess=extract_component)
    
    data_xrsel = data_xrsel.rename({'lon':'longitude','lat':'latitude'})

    #convert cm to m
    if data_xrsel['amplitude'].attrs['units'] == 'cm':
        data_xrsel['amplitude'] /= 100
        data_xrsel['amplitude'].attrs['units'] = 'm'
    
    # compute imag/real components to avoid zero-crossing interpolation issues with phase
    if 'wl_real' not in data_xrsel.data_vars:
        data_xrsel_phs_rad = np.deg2rad(data_xrsel['phase'])
        data_xrsel['wl_real'] = data_xrsel['amplitude'] * np.cos(data_xrsel_phs_rad)
        data_xrsel['wl_imag'] = data_xrsel['amplitude'] * np.sin(data_xrsel_phs_rad)
    
    data_xrsel['compnames'] = xr.DataArray(component_list,dims=('compno'))
    data_xrsel = data_xrsel.set_index({'compno':'compnames'})
    
    data_interp = interp_regularnc_to_plipointsDataset(data_xr_reg=data_xrsel, gdf_points=gdf_points, load=load)
    data_interp['phase_new'] = np.rad2deg(np.arctan2(data_interp['wl_imag'],data_interp['wl_real']))
    return data_interp


def check_time_extent(data_xr, tstart, tstop):
    #convert tstart/tstop from str/dt.datetime/pd.Timestamp to pd.Timestamp. WARNING: when supplying '05-04-2016', monthfirst is assumed, so 2016-05-04 will be the result (may instead of april). So always supply yyyy-mm-dd datestrings
    tstart = pd.Timestamp(tstart)
    tstop = pd.Timestamp(tstop)
    
    #get timevar and compare requested dates
    timevar = data_xr['time']
    xr_tstartstop = pd.to_datetime(timevar.isel(time=[0,-1]).to_series())
    nc_tstart = xr_tstartstop.index[0]
    nc_tstop = xr_tstartstop.index[-1]
    if tstart < nc_tstart:
        raise OutOfRangeError(f'requested tstart {tstart} outside of available range {nc_tstart} to {nc_tstop}.')
    if tstop > nc_tstop:
        raise OutOfRangeError(f'requested tstop {tstop} outside of available range {nc_tstart} to {nc_tstop}.')


def ds_apply_conventions(data_xr):
    ncbnd_construct = get_ncbnd_construct()
    dimn_depth = ncbnd_construct['dimn_depth']
    varn_depth = ncbnd_construct['varn_depth']
    
    #rename dims and vars time/depth/lat/lon/x/y #TODO: this has to be phased out some time, or made as an argument or merged with conversion_dict?
    rename_dims_dict = {'depth':dimn_depth, # depth for CMEMS and many others
                        'lev':dimn_depth, # lev for GFDL
                        'lon':'longitude','lat':'latitude'}
    for k,v in rename_dims_dict.items():
        if k in data_xr.dims and v not in data_xr.dims:
            data_xr = data_xr.rename_dims({k:v})
            print(f'dimension {k} renamed to {v}')
        if k in data_xr.variables and v not in data_xr.variables:
            data_xr = data_xr.rename_vars({k:v})
            print(f'varname {k} renamed to {v}')
    
    # make depth positive up if not yet the case
    if varn_depth in data_xr.coords:
        if 'positive' in data_xr[varn_depth].attrs.keys():
            if data_xr[varn_depth].attrs['positive'] == 'down': #attribute appears in CMEMS, GFDL and CMCC, save to assume presence?
                data_xr[varn_depth] = -data_xr[varn_depth]
                data_xr[varn_depth].attrs['positive'] = 'up'
    
    # get calendar and maybe convert_calendar, makes sure that nc_tstart/nc_tstop are of type pd._libs.tslibs.timestamps.Timestamp
    data_xr_calendar = data_xr['time'].dt.calendar
    if data_xr_calendar != 'proleptic_gregorian': #this is for instance the case in case of noleap (or 365_days) calendars from GFDL and CMCC
        units_copy = data_xr['time'].encoding['units']
        print(f'WARNING: calendar different than proleptic_gregorian found ({data_xr_calendar}), convert_calendar is called so check output carefully. It should be no issue for datasets with a monthly interval.')
        data_xr = data_xr.convert_calendar('standard') #TODO: does this not result in 29feb nan values in e.g. GFDL model? Check missing argument at https://docs.xarray.dev/en/stable/generated/xarray.Dataset.convert_calendar.html
        data_xr['time'].encoding['units'] = units_copy #put back dropped units
    
    # 360 to 180 conversion
    if 'longitude' in data_xr.variables:
        convert_360to180 = (data_xr['longitude'].to_numpy()>180).any() #TODO: replace to_numpy() with load()
        if convert_360to180: #TODO: make more flexible for models that eg pass -180/+180 crossing (add overlap at lon edges).
            data_xr.coords['longitude'] = (data_xr.coords['longitude'] + 180) % 360 - 180
            data_xr = data_xr.sortby(data_xr['longitude'])
    
    return data_xr


def ds_apply_conversion_dict(data_xr, conversion_dict, quantity):
    # rename variables from conversion_dict
    for k,v in conversion_dict.items():
        ncvarn = v['ncvarname']
        if ncvarn in data_xr.variables.mapping.keys() and k == quantity: #k in quantity_list so phyc is not always renamed to tracerbndPON1 (first in conversion_dict)
            data_xr = data_xr.rename({ncvarn:k})
            print(f'variable {ncvarn} renamed to {k}')

    # optional conversion of units. Multiplications or other simple operatiors do not affect performance (dask.array(getitem) becomes dask.array(mul). With more complex operation it is better do do it on the interpolated array.
    # TODO: maybe do unit conversion before interp or is that computationally heavy?
    if 'conversion' in conversion_dict[quantity].keys(): #if conversion is present, unit key must also be in conversion_dict
        print(f'> converting units from [{data_xr[quantity].attrs["units"]}] to [{conversion_dict[quantity]["unit"]}]')
        # conversion drops all attributes of which units (which are changed anyway)
        data_xr[quantity] = data_xr[quantity] * conversion_dict[quantity]['conversion']
        # add unit attribute with resulting unit
        quantity_attrs = {'units': conversion_dict[quantity]['unit']}
        data_xr[quantity] = data_xr[quantity].assign_attrs(quantity_attrs)
    return data_xr


def open_prepare_dataset(dir_pattern, quantity, tstart, tstop, conversion_dict=None, refdate_str=None, chunks=None):
    """
    empty docstring
    """
    
    if conversion_dict is None:
        conversion_dict = get_conversion_dict()
    
    file_list_nc = glob.glob(str(dir_pattern))
    print(f'loading mfdataset of {len(file_list_nc)} files with pattern(s) {dir_pattern}')
    
    data_xr = xr.open_mfdataset(file_list_nc, chunks=chunks, join="exact") #TODO: does chunks argument solve "PerformanceWarning: Slicing is producing a large chunk."? {'time':1} is not a convenient chunking to use for timeseries extraction
    
    data_xr = ds_apply_conventions(data_xr=data_xr)
    data_xr = ds_apply_conversion_dict(data_xr=data_xr, conversion_dict=conversion_dict, quantity=quantity)

    # retrieve var(s) (after potential longitude conversion)
    data_xr_vars = data_xr[[quantity]]
    
    # slice dataset to times outside of requested time range
    data_xr_vars = _ds_sel_time_outside(
        ds=data_xr_vars,
        tstart=tstart,
        tstop=tstop,
        )
        
    #optional refdate changing
    if refdate_str is not None:
        if 'long_name' in data_xr_vars.time.attrs: #for CMEMS it is 'hours since 1950-01-01', which would be wrong now #TODO: consider also removing attrs for depth/varname, since we would like to have salinitybnd/waterlevel instead of Salinity/sea_surface_height in xr plots?
            del data_xr_vars.time.attrs['long_name']
        data_xr_vars.time.encoding['units'] = refdate_str
    
    return data_xr_vars


def _ds_sel_time_outside(ds: xr.Dataset, tstart, tstop) -> xr.Dataset:
    """
    Subset the dataset on time, making sure the requested times are always
    included. If there is no exact match for start/end times, the previous/next
    timestamp is taken as an extreme.
    Inspired by copernicusmarine.download_functions.subset_xarray.py
    """    
    check_time_extent(ds, tstart, tstop)
    external_minimum = ds.sel(time=tstart, method="pad")
    external_maximum = ds.sel(time=tstop, method="backfill")
    time_slice = slice(external_minimum.time.values, external_maximum.time.values)
    ds_sel = ds.sel(time=time_slice) 
    return ds_sel


def interp_regularnc_to_plipointsDataset(data_xr_reg, gdf_points, load=True):
    
    ncbnd_construct = get_ncbnd_construct()
    varn_pointx = ncbnd_construct['varn_pointx']
    varn_pointy = ncbnd_construct['varn_pointy']
    
    da_plipoints = da_from_gdf_points(gdf_points)
    
    #interpolation to lat/lon combinations
    print('> interp mfdataset to all PolyFile points (lat/lon coordinates)')
    
    # linear, nearest, combine_first
    data_interp_lin = data_xr_reg.interp(longitude=da_plipoints[varn_pointx], latitude=da_plipoints[varn_pointy],
                                         method='linear', 
                                         kwargs={'bounds_error':True}) #error is only raised upon load(), so when the actual value retrieval happens
    data_interp_near = data_xr_reg.interp(longitude=da_plipoints[varn_pointx], latitude=da_plipoints[varn_pointy],
                                          method='nearest', 
                                          kwargs={'bounds_error':True}) #error is only raised upon load(), so when the actual value retrieval happens
    data_interp = data_interp_lin.combine_first(data_interp_near)
    
    # drop original latitude/longitude vars (lat/lon in da_plipoints)
    data_interp = data_interp.drop_vars(['latitude','longitude'])
    
    if not load:
        return data_interp
    
    print(f'> actual extraction of data from netcdf with .load() (for {len(gdf_points)} plipoints at once, this might take a while)')
    dtstart = dt.datetime.now()
    try:
        data_interp_loaded = data_interp.load() #loading data for all points at once is more efficient compared to loading data per point in loop 
    except ValueError: #generate a proper error with outofbounds requested coordinates, default is "ValueError: One of the requested xi is out of bounds in dimension 0" #TODO: improve error in xarray
        lonvar_vals = data_xr_reg['longitude'].to_numpy()
        latvar_vals = data_xr_reg['latitude'].to_numpy()
        data_pol_pd = data_interp[[varn_pointx,varn_pointy]].to_dataframe()
        bool_reqlon_outbounds = (data_pol_pd[varn_pointx] <= lonvar_vals.min()) | (data_pol_pd[varn_pointx] >= lonvar_vals.max())
        bool_reqlat_outbounds = (data_pol_pd[varn_pointy] <= latvar_vals.min()) | (data_pol_pd[varn_pointy] >= latvar_vals.max())
        reqlatlon_pd = pd.DataFrame({'longitude':data_pol_pd[varn_pointx],'latitude':data_pol_pd[varn_pointy],'lon outbounds':bool_reqlon_outbounds,'lat outbounds':bool_reqlat_outbounds})
        reqlatlon_pd_outbounds = reqlatlon_pd.loc[bool_reqlon_outbounds | bool_reqlat_outbounds]
        raise ValueError(f'{len(reqlatlon_pd_outbounds)} of requested pli points are out of bounds (valid longitude range {lonvar_vals.min()} to {lonvar_vals.max()}, valid latitude range {latvar_vals.min()} to {latvar_vals.max()}):\n{reqlatlon_pd_outbounds}')
    time_passed = (dt.datetime.now()-dtstart).total_seconds()
    print(f'>>time passed: {time_passed:.2f} sec')

    return data_interp_loaded


def interp_uds_to_plipoints(uds:xu.UgridDataset, gdf:geopandas.GeoDataFrame) -> xr.Dataset:
    """
    To interpolate an unstructured dataset (like a _map.nc file) read with xugrid to plipoint locations
    
    Parameters
    ----------
    uds : xu.UgridDataset
        dfm model output read using dfm_tools. Dims: mesh2d_nLayers, mesh2d_nInterfaces, time, mesh2d_nNodes, mesh2d_nFaces, mesh2d_nMax_face_nodes, mesh2d_nEdges.
    gdf : geopandas.GeoDataFrame
        gdf with location, geometry (Point) and crs corresponding to model crs.

    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    ds : TYPE
        xr.Dataset with dims: node, time, z.

    """
    # TODO: this function requires gdf with points, but the interp_regularnc_to_plipoints requires paths to plifiles (and others also)
    
    facedim = uds.grid.face_dimension
    edgedim = uds.grid.edge_dimension
    nodedim = uds.grid.node_dimension
    ncbnd_construct = get_ncbnd_construct()
    dimn_point = ncbnd_construct['dimn_point']
    varn_pointname = ncbnd_construct['varn_pointname']
    
    # drop node/edge variables since they are not interpolated but 
    # get an additional face dimension if some points are out of bounds
    # TODO: revert after fixing https://github.com/Deltares/xugrid/issues/274
    vars_without_facedim = []
    for varn in uds.variables:
        if facedim not in uds.variables[varn].dims:
            vars_without_facedim.append(varn)
    uds_face = uds.drop_vars(vars_without_facedim)
    
    # interpolate to provided points
    ds = uds_face.ugrid.sel_points(x=gdf.geometry.x, y=gdf.geometry.y)
    
    # re-add removed variables again, sometimes important for e.g. depth
    # TODO: remove after fixing https://github.com/Deltares/xugrid/issues/274
    for varn in vars_without_facedim:
        vardims = uds.variables[varn].dims
        if edgedim not in vardims and nodedim not in vardims:
            ds[varn] = uds.variables[varn]
    
    # rename station dimname and varname (is index, are both mesh2d_nFaces to start with)
    ds = ds.rename({facedim:dimn_point}) # rename mesh2d_nFaces to plipoints
    ds = ds.rename_vars({dimn_point:varn_pointname})
    # change name of plipoint (node to gdf name)
    ds[varn_pointname] = xr.DataArray(gdf[varn_pointname].tolist(), dims=dimn_point)
    return ds


def interp_hisnc_to_plipoints(data_xr_his, file_pli, kdtree_k=3, load=True):
    """
    interpolate stations in hisfile to points of polyfile
    """
    #KDTree, finds minimal eucledian distance between points (haversine would be better). Alternatively sklearn.neighbors.BallTree: tree = BallTree(data_lonlat_pd)
    
    ncbnd_construct = get_ncbnd_construct()
    dimn_point = ncbnd_construct['dimn_point']
    
    datavars_list = list(data_xr_his.data_vars)
    if len(datavars_list)>5:
        print('WARNING: more than 5 data_vars, you might want to subset your data_xr_his')
    #read hisfile and make KDTree of xy of hisstations
    hisstations_pd = data_xr_his.stations.to_dataframe() #TODO: add check if stations are strings? (whether preprocess_hisnc was used)
    tree_nest2 = KDTree(hisstations_pd[['station_x_coordinate','station_y_coordinate']])
    
    #read polyfile
    polyfile_object = hcdfm.PolyFile(file_pli)
    gdf_points = PolyFile_to_geodataframe_points(polyfile_object)
    gdf_coords = gdf_points.get_coordinates()
    
    # query k nearest hisstations (names)
    plicoords_distance2, plicoords_nestpointidx = tree_nest2.query(gdf_coords, k=kdtree_k)
    da_plicoords_nestpointidx = xr.DataArray(plicoords_nestpointidx, dims=(dimn_point,'nearestkpoints'))
    da_plicoords_nestpointnames = data_xr_his.stations.isel(stations=da_plicoords_nestpointidx)
    
    # convert gdf to xarray dataset
    data_interp = da_from_gdf_points(gdf_points)
    
    #interpolate hisfile variables to plipoints
    for varone in datavars_list:
        da_dist = xr.DataArray(plicoords_distance2, dims=(dimn_point,'nearestkpoints'))
        da_invdistweight = (1/da_dist)/(1/da_dist).sum(dim='nearestkpoints') #TODO: set weigths for invalid points to 0 (and increase others weights so sum remains 1)
        data_xr_his_pli_knearest = data_xr_his[varone].sel(stations=da_plicoords_nestpointnames)
        data_interp[varone] = (data_xr_his_pli_knearest * da_invdistweight).sum(dim='nearestkpoints')
        data_interp[varone].attrs = data_xr_his[varone].attrs #copy units and other attributes
        
    if not load:
        return data_interp
    
    print('loading dataset (might take a while, but increases performance of point loop significantly)')
    data_interp = data_interp.load()
    print('done')

    return data_interp
    

def plipointsDataset_to_ForcingModel(plipointsDataset):
    """
    empty docstring
    """
    
    ncbnd_construct = get_ncbnd_construct()
    dimn_point = ncbnd_construct['dimn_point']
    dimn_depth = ncbnd_construct['dimn_depth']
    varn_pointname = ncbnd_construct['varn_pointname']
    
    plipointsDataset = maybe_convert_fews_to_dfmt(plipointsDataset)
    
    quantity_list = list(plipointsDataset.data_vars)
    npoints = len(plipointsDataset[dimn_point])
    
    #start conversion to Forcingmodel object
    print(f'Converting {npoints} plipoints to hcdfm.ForcingModel():',end='')
    dtstart = dt.datetime.now()
    ForcingModel_object = hcdfm.ForcingModel()
    for iP in range(npoints):
        print(f' {iP+1}',end='')
        
        #select data for this point, ffill nans, concatenating time column, constructing T3D/TimeSeries and append to hcdfm.ForcingModel()
        datablock_xr_onepoint = plipointsDataset.isel({dimn_point:iP})
        plipoint_name = str(datablock_xr_onepoint[varn_pointname].to_numpy())
        plipoint_onlynan = False
        for quan in quantity_list:
            datablock_xr_onepoint[quan].attrs['locationname'] = plipoint_name #TODO: is there a nicer way of passing this data?
            if datablock_xr_onepoint[quan].isnull().all(): # check if all values of plipoint are nan (on land)
                plipoint_onlynan = True
                logger.warning(f'Plipoint "{plipoint_name}" might be on land since it only contain nan values. '
                               'This point is skipped to avoid bc-writing errors. Consider altering your PolyFile or extrapolate the data.')
        
        # skip this point if quantity has only-nan values
        if plipoint_onlynan:
            continue
        
        if dimn_depth in plipointsDataset.dims:
            ts_one = Dataset_to_T3D(datablock_xr_onepoint)
        elif 'amplitude' in quantity_list:
            ts_one = Dataset_to_Astronomic(datablock_xr_onepoint)
        else:
            ts_one = Dataset_to_TimeSeries(datablock_xr_onepoint)
        ForcingModel_object.forcing.append(ts_one)
    print(f'. >> done in {(dt.datetime.now()-dtstart).total_seconds():.2f} sec')
    
    return ForcingModel_object




import os
import pandas as pd
from pathlib import Path
import xarray as xr
from dfm_tools.errors import OutOfRangeError
from dfm_tools.interpolate_grid2bnd import _ds_sel_time_outside
import cdsapi
import copernicusmarine
from copernicusmarine.core_functions.credentials_utils import InvalidUsernameOrPassword
import cftime
import getpass
from requests import HTTPError
from requests.exceptions import SSLError
import logging

__all__ = [
    "download_ERA5",
    "download_CMEMS",
    "download_OPeNDAP",
]

logger = logging.getLogger(__name__)

COPERNICUSMARINE_OPTIMIZE_ARGS = dict(
    # speed up copernicusmarine.open_dataset() with the following arguments
    # this optimizes chunking for downloading with daily frequency
    # https://github.com/Deltares/dfm_tools/issues/1033
    # also relevant to get time bounds
    # https://github.com/Deltares/dfm_tools/issues/1058
    service="arco-geo-series",
    chunk_size_limit=0,
    # prevent the need for buffering time/spatial extent
    # https://github.com/Deltares/dfm_tools/issues/1050
    coordinates_selection_method='outside',
)


def download_ERA5(varkey,
                  longitude_min, longitude_max, latitude_min, latitude_max, 
                  date_min, date_max,
                  dir_output='.', overwrite=False):
    """
    empty docstring
    """
    
    #TODO: describe something about the .cdsapirc file
    #TODO: make this function cdsapi generic, instead of ERA5 hardcoded (make flexible for product_type/name/name_output) (variables_dict is not used actively anymore, so this is possible)
    
    # create ~/.cdsapirc if it does not exist
    cds_credentials()
    
    # import cdsapi and create a Client instance # https://cds.climate.copernicus.eu/api-how-to
    c = cdsapi.Client()

    #dictionary with ERA5 variables #this is not actively used
    variables_dict = {'ssr':'surface_net_solar_radiation',
                      'sst':'sea_surface_temperature',
                      'strd':'surface_thermal_radiation_downwards',
                      'slhf':'surface_latent_heat_flux',
                      'sshf':'surface_sensible_heat_flux',
                      'str':'surface_net_thermal_radiation',
                      'chnk':'charnock',
                      'd2m':'2m_dewpoint_temperature',
                      't2m':'2m_temperature',
                      'tcc':'total_cloud_cover',
                      'msl':'mean_sea_level_pressure',
                      'u10':'10m_u_component_of_wind',
                      'u10n':'10m_u_component_of_neutral_wind',
                      'v10':'10m_v_component_of_wind',
                      'v10n':'10m_v_component_of_neutral_wind',
                      # mer and mtpr are now called avg_ie and avg_tprate
                      # this is reverted in dfmt.preprocess_ERA5() after fixing
                      # https://github.com/Deltares/dfm_tools/issues/1140
                      'mer':'mean_evaporation_rate', 
                      'mtpr':'mean_total_precipitation_rate',
                      'rhoao':'air_density_over_the_oceans',
                      }
    if varkey not in variables_dict.keys(): #TODO: how to get list of available vars? mean_sea_level_pressure and msl both return a dataset with msl varkey, but standard_name air_pressure_at_mean_sea_level returns an error
        raise KeyError(f'"{varkey}" not available, choose from: {", ".join(variables_dict.keys())}')
    
    period_range = pd.period_range(date_min,date_max,freq='M')
    if len(period_range) == 0:
        raise ValueError(f"requested time extents ({date_min} to {date_max}) "
                         "resulted in empty period_range")
    print(f'retrieving data from {period_range[0]} to {period_range[-1]} (freq={period_range.freq})')
    
    #make sure the data fully covers the desired spatial extent. Download 1 additional grid cell (resolution is 1/4 degrees) in both directions
    buffer = 2/4
    longitude_min -= buffer
    longitude_max += buffer
    latitude_min  -= buffer
    latitude_max  += buffer
    
    for date in period_range:
        name_output = f'era5_{varkey}_{date.strftime("%Y-%m")}.nc'
        file_out = Path(dir_output,name_output)
        if file_out.is_file() and not overwrite:
            print(f'"{name_output}" found and overwrite=False, continuing.')
            continue
        print (f'retrieving ERA5 data for variable "{varkey}" and month {date.strftime("%Y-%m")} (YYYY-MM)')

        request_dict = {'product_type':'reanalysis',
                        'variable':variables_dict[varkey],
                        'year':date.strftime('%Y'),
                        'month':date.strftime('%m'),
                        #'month':[f'{x:02d}' for x in range(1,12+1)], #all months, but instead retrieving per month
                        'day':[f'{x:02d}' for x in range(1,31+1)], #all days
                        'time':[f'{x:02d}:00' for x in range(0,23+1)], #all times/hours
                        'area':[latitude_max,longitude_min,latitude_min,longitude_max], # north, west, south, east. default: global - option not available through the Climate Data Store (CDS) web interface (for cmip data)
                        #'grid': [1.0, 1.0], # latitude/longitude grid: east-west (longitude) and north-south resolution (latitude). default: 0.25 x 0.25 - option not available through the Climate Data Store (CDS) web interface
                        "data_format": "netcdf",
                        "download_format": "unarchived",
                        }
        
        resp = c.retrieve(
            name='reanalysis-era5-single-levels', 
            request=request_dict,
            )
        resp.download(target=file_out)


def cds_credentials():
    """
    get cdsapikey from environment variables or file or query via getpass if necessary
    """
    cds_url_default = "https://cds.climate.copernicus.eu/api"
    
    # TODO: put this in a PR at https://github.com/ecmwf/cdsapi
    if "CDSAPI_KEY" in os.environ.keys():
        # in case of envvars, also set CDSAPI_URL envvar so it does not have to be supplied
        cds_url = os.environ.get("CDSAPI_URL", cds_url_default)
        os.environ["CDSAPI_URL"] = cds_url
    
    try:
        # gets url/key from env vars or ~/.cdsapirc file
        c = cdsapi.Client()
    except Exception as e:
        if "Missing/incomplete configuration file" in str(e):
            # query apikey if not present in file or envvars
            print("Downloading CDS/ERA5 data requires a ECMWF API-key, copy "
                  "your API-key from https://cds.climate.copernicus.eu/profile "
                  "(first register, login and accept the terms). "
                  "More info in https://forum.ecmwf.int/t/3743.")
            cds_url = cds_url_default
            cds_apikey = getpass.getpass("\nEnter your ECMWF API-key (string with dashes): ")
            cds_set_credentials(cds_url, cds_apikey)
            c = cdsapi.Client()
        else:
            raise e
    
    # check if the authentication works and catch some exceptions that occur
    # with old or incorrect cdsapi urls or keys
    try:
        # c = cdsapi.Client()
        c.client.check_authentication()
        print('found ECMWF API-key and authorization successful')
    except AttributeError as e:
        #"AttributeError: 'Client' object has no attribute 'client'"
        reason = str(e) + ". This means an old CDSAPI KEY was found (with :)"
        cds_remove_credentials_raise(reason=reason)
    except (HTTPError, SSLError) as e:
        reason = str(e)
        if "SSLError" in str(e):
            # SSLError: [SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: 
            # self-signed certificate (_ssl.c:1000)
            reason += ". This means an old CDSAPI URL was found (with cds-beta)"
        elif "404 Client Error: Not Found for url:" in str(e):
            # HTTPError: 404 Client Error: Not Found for url: https://cds.climate.
            # copernicus.eu/api/v2/profiles/v1/account/verification/pat
            reason += "This means an old/incorrect CDSAPI URL was found (e.g. with v2)"
        cds_remove_credentials_raise(reason=reason)


def cds_get_file():
    file_cds_credentials = os.environ.get("CDSAPI_RC", os.path.expanduser("~/.cdsapirc"))
    return file_cds_credentials

    
def cds_set_credentials_rcfile(cds_url, cds_apikey):
    file_cds_credentials = cds_get_file()
    with open(file_cds_credentials,'w') as fc:
        fc.write(f'url: {cds_url}\n')
        fc.write(f'key: {cds_apikey}')
    
    
def cds_set_credentials(cds_url, cds_apikey):
    # set env vars
    os.environ["CDSAPI_URL"] = cds_url
    os.environ["CDSAPI_KEY"] = cds_apikey
    
    # set ~/.cdsapirc file
    cds_set_credentials_rcfile(cds_url, cds_apikey)


def cds_remove_credentials_raise(reason=''):
    """
    remove CDS url and uid:apikey environment variables and ~/.cdsapirc file
    environment variables defined in https://github.com/ecmwf/cdsapi/blob/main/cdsapi/api.py
    """
    
    keys_toremove = ["CDSAPI_URL",
                     "CDSAPI_KEY",
                     "CDSAPI_RC"]
    for key in keys_toremove:
        if key in os.environ.keys():
            os.environ.pop(key)

    file_cds_credentials = cds_get_file()
    if os.path.isfile(file_cds_credentials):
        os.remove(file_cds_credentials)
    
    raise ValueError(f"{reason}. The CDS/ECMWF apikey file (~/.cdsapirc) was deleted. "
                     "Please try again and provide your apikey when prompted.")


def download_CMEMS(varkey,
                   longitude_min, longitude_max, latitude_min, latitude_max, 
                   date_min, date_max, freq='D',
                   dataset_id=None,
                   dir_output='.', file_prefix='', overwrite=False):
    """
    Download CMEMS data by providing varkey and optional dataset_id. If no
    dataset_id is provided, it is automatically derived based on the varkey
    and the requested period. For daily mean data, the dataset is automatically
    shifted with 12 hours to convert start-of-interval to center-of-interval
    timestamps.
    
    The requested spatial and time extents are automatically buffered to make
    sure all requested values are included in the returned dataset. This
    behaviour is more inclusive than the default xarray.Dataset.sel() method.
    
    The data is downloaded in files per day or month, depending on the freq
    argument. When downloading monthly or yearly means make sure not to request
    a higher freq, since that would result in many empty files.
    
    More info about copernicusmarine toolbox available at:
    https://help.marine.copernicus.eu/en/articles/8283072-copernicus-marine-toolbox-api-subset
    
    """
    copernicusmarine_credentials()
    
    if dataset_id is None:
        dataset_id = copernicusmarine_get_dataset_id(varkey, date_min, date_max)
    
    # date_range with same start as stoptime is a bit tricky so we limit freqs
    # https://github.com/Deltares/dfm_tools/issues/720
    if freq not in ["D","M","Y"]:
        raise ValueError(f"freq should be D/M/Y, not {freq}")
    # prevent daily freq for datasets with monthly intervals
    # to avoid downloading empty files
    if "_P1M" in dataset_id and freq=="D":
        freq = "M"
    print(f"downloading {varkey} from {dataset_id} with freq={freq}")
    
    dataset = copernicusmarine.open_dataset(
         dataset_id = dataset_id,
         variables = [varkey],
         minimum_longitude = longitude_min,
         maximum_longitude = longitude_max,
         minimum_latitude = latitude_min,
         maximum_latitude = latitude_max,
         **COPERNICUSMARINE_OPTIMIZE_ARGS,
    )
    
    # possibly shift times with 12 hours (to center-of-interval times)
    dataset = copernicusmarine_dataset_timeshift(ds=dataset, dataset_id=dataset_id)
    
    # slice to outside times after opening dataset and correcting times
    dataset = _ds_sel_time_outside(ds=dataset, tstart=date_min, tstop=date_max)
    
    Path(dir_output).mkdir(parents=True, exist_ok=True)
    
    # get time extent from dataset, it can be different than requested
    # due to coordinates_selection_method='outside'
    start_datetime = dataset.time.isel(time=0).values
    end_datetime = dataset.time.isel(time=-1).values
    period_range = pd.period_range(start_datetime, end_datetime, freq=freq)
    for date in period_range:
        date_str = str(date)
        name_output = f'{file_prefix}{varkey}_{date_str}.nc'
        output_filename = Path(dir_output, name_output)
        if output_filename.is_file() and not overwrite:
            print(f'"{name_output}" found and overwrite=False, continuing.')
            continue
        dataset_perperiod = dataset.sel(time=slice(date_str, date_str))
        print(f'writing netcdf file: {name_output}: ',end='')
        dtstart = pd.Timestamp.now()
        dataset_perperiod.to_netcdf(output_filename)
        print(f'{(pd.Timestamp.now()-dtstart).total_seconds():.2f} sec')


def copernicusmarine_get_product(date_min, date_max, vartype):
    """
    retrieve the time extents of reanalysis, reanalysis-interim and forecast
    products. The time extents of phy and bio can be different, so they have
    to be retrieved for both.
    """
    
    assert vartype in ['phy','bio']
    
    # convert requested times to pandas timestamps for >= and <= comparison
    date_min = pd.Timestamp(date_min)
    date_max = pd.Timestamp(date_max)
    
    # time extents as global variables, so they only has to be retreived once per download run (otherwise once per variable)
    global phy_reanalysis_tstart, phy_reanalysis_tstop
    global phy_reanalysis_int_tstart, phy_reanalysis_int_tstop
    global phy_forecast_tstart, phy_forecast_tstop
    global bio_reanalysis_tstart, bio_reanalysis_tstop
    global bio_reanalysis_int_tstart, bio_reanalysis_int_tstop
    global bio_forecast_tstart, bio_forecast_tstop
    
    # retrieve times
    if vartype=='phy' and 'phy_reanalysis_tstart' not in globals():
        print('retrieving time range of CMEMS reanalysis, reanalysis-interim and forecast products (phy)') #assuming here that physchem and bio reanalyisus/multiyear datasets have the same enddate, this seems safe
        phy_reanalysis_tstart, phy_reanalysis_tstop = copernicusmarine_dataset_timerange(dataset_id="cmems_mod_glo_phy_my_0.083deg_P1D-m")
        phy_reanalysis_int_tstart, phy_reanalysis_int_tstop = copernicusmarine_dataset_timerange(dataset_id="cmems_mod_glo_phy_myint_0.083deg_P1D-m")
        phy_forecast_tstart, phy_forecast_tstop = copernicusmarine_dataset_timerange(dataset_id="cmems_mod_glo_phy_anfc_0.083deg_P1D-m")
    if vartype=='bio' and 'bio_reanalysis_tstart' not in globals():
        print('retrieving time range of CMEMS reanalysis, reanalysis-interim and forecast products (bio)') #assuming here that physchem and bio reanalyisus/multiyear datasets have the same enddate, this seems safe
        bio_reanalysis_tstart, bio_reanalysis_tstop = copernicusmarine_dataset_timerange(dataset_id="cmems_mod_glo_bgc_my_0.25deg_P1D-m")
        bio_reanalysis_int_tstart, bio_reanalysis_int_tstop = copernicusmarine_dataset_timerange(dataset_id="cmems_mod_glo_bgc_myint_0.25deg_P1D-m")
        bio_forecast_tstart, bio_forecast_tstop = copernicusmarine_dataset_timerange(dataset_id="cmems_mod_glo_bgc-pft_anfc_0.25deg_P1D-m")
    
    # set current start/stop times dependent on whether we request phy/bio
    if vartype=='phy':
        reanalysis_tstart, reanalysis_tstop = phy_reanalysis_tstart, phy_reanalysis_tstop
        reanalysis_int_tstart, reanalysis_int_tstop = phy_reanalysis_int_tstart, phy_reanalysis_int_tstop
        forecast_tstart, forecast_tstop = phy_forecast_tstart, phy_forecast_tstop
    if vartype=='bio' :
        reanalysis_tstart, reanalysis_tstop = bio_reanalysis_tstart, bio_reanalysis_tstop
        reanalysis_int_tstart, reanalysis_int_tstop = bio_reanalysis_int_tstart, bio_reanalysis_int_tstop
        forecast_tstart, forecast_tstop = bio_forecast_tstart, bio_forecast_tstop
        
    if (date_min >= reanalysis_tstart) & (date_max <= reanalysis_tstop):
        product = 'reanalysis'
    elif (date_min >= reanalysis_int_tstart) & (date_max <= reanalysis_int_tstop):
        product = 'reanalysis-interim'
    elif (date_min >= forecast_tstart) & (date_max <= forecast_tstop):
        product = 'analysisforecast'
    else:
        raise ValueError(f'The requested timerange ({date_min} to {date_max}) is not fully within the timerange of one of the following datasets:\n'
                         f'- reanalysis: {reanalysis_tstart} to {reanalysis_tstop}\n'
                         f'- reanalysis-interim: {reanalysis_int_tstart} to {reanalysis_int_tstop}\n'
                         f'- analysisforecast: {forecast_tstart} to {forecast_tstop}\n'
                         'Please adjust the requested timerange and try again.')
    print(f"The CMEMS '{product}' product will be used.")
    return product


def copernicusmarine_get_dataset_id(varkey, date_min, date_max):
    #TODO: maybe get dataset_id from 'copernicusmarine describe --include-datasets --contains <search_token>'
    
    if varkey in ['bottomT','tob','mlotst','siconc','sithick','so','thetao','uo','vo','usi','vsi','zos']: #for physchem
        vartype = 'phy'
    elif varkey in ['nppv','o2','talk','dissic','ph','spco2','no3','po4','si','fe','chl','phyc']: # for bio
        vartype = 'bio'
    else:
        raise KeyError(f"unknown varkey for cmems: {varkey}")
    
    product = copernicusmarine_get_product(date_min, date_max, vartype)
    
    if vartype == 'phy': #for physchem
        # resolution is 1/12 degrees in lat/lon dimension, but a bit more/less in alternating cells
        if product == 'analysisforecast': # forecast: https://data.marine.copernicus.eu/product/GLOBAL_ANALYSISFORECAST_PHY_001_024/description
            if varkey in ['uo','vo']: # anfc datset is splitted over multiple urls
                dataset_id = 'cmems_mod_glo_phy-cur_anfc_0.083deg_P1D-m'
            elif varkey in ['so']:
                dataset_id = 'cmems_mod_glo_phy-so_anfc_0.083deg_P1D-m'
            elif varkey in ['thetao']:
                dataset_id = 'cmems_mod_glo_phy-thetao_anfc_0.083deg_P1D-m'
            else:
                dataset_id = 'cmems_mod_glo_phy_anfc_0.083deg_P1D-m'
        elif product == 'reanalysis-interim': # reanalysis-interim: https://data.marine.copernicus.eu/product/GLOBAL_MULTIYEAR_BGC_001_029/description
            dataset_id = 'cmems_mod_glo_phy_myint_0.083deg_P1D-m'
        else: # reanalysis: https://data.marine.copernicus.eu/product/GLOBAL_MULTIYEAR_PHY_001_030/description
            dataset_id = 'cmems_mod_glo_phy_my_0.083deg_P1D-m'
    elif vartype == 'bio': # for bio
        # resolution is 1/4 degrees
        if product == 'analysisforecast': # forecast: https://data.marine.copernicus.eu/product/GLOBAL_ANALYSISFORECAST_BGC_001_028/description
            if varkey in ['nppv','o2']:
                dataset_id = 'cmems_mod_glo_bgc-bio_anfc_0.25deg_P1D-m'
            elif varkey in ['talk','dissic','ph']:
                dataset_id = 'cmems_mod_glo_bgc-car_anfc_0.25deg_P1D-m'
            elif varkey in ['spco2']:
                dataset_id = 'cmems_mod_glo_bgc-co2_anfc_0.25deg_P1D-m'
            elif varkey in ['no3','po4','si','fe']:
                dataset_id = 'cmems_mod_glo_bgc-nut_anfc_0.25deg_P1D-m'
            elif varkey in ['chl','phyc']:
                dataset_id = 'cmems_mod_glo_bgc-pft_anfc_0.25deg_P1D-m'
        elif product == 'reanalysis-interim': # reanalysis-interim: https://data.marine.copernicus.eu/product/GLOBAL_MULTIYEAR_BGC_001_029/description
            dataset_id = 'cmems_mod_glo_bgc_myint_0.25deg_P1D-m'
        else: # reanalysis: https://data.marine.copernicus.eu/product/GLOBAL_MULTIYEAR_BGC_001_029/description
            if varkey in ['chl', 'no3', 'nppv', 'o2', 'po4', 'si']:
                dataset_id = 'cmems_mod_glo_bgc_my_0.25deg_P1D-m'
            else:
                # phyc/fe/ph/spco2 are not available in daily mean dataset
                # use monthly mean dataset instead
                # https://github.com/Deltares/dfm_tools/issues/622
                dataset_id = 'cmems_mod_glo_bgc_my_0.25deg_P1M-m'
    else:
        raise ValueError(f"unknown vartype for cmems: {vartype}")
    return dataset_id


def copernicusmarine_credentials():
    """
    Login at copernicusmarine if user not logged in yet.
    Works via prompt, environment variables or credentials file.
    """
    # first check whether (valid) credentials are already present
    success = copernicusmarine.login(check_credentials_valid=True)
    if success:
        return

    # call the login function if no (valid) credentials present
    success = copernicusmarine.login()

    # raise if credentials are incorrect
    if not success:
        raise InvalidUsernameOrPassword("Invalid credentials, please try again")


def copernicusmarine_dataset_timeshift(ds, dataset_id):
    """
    correct daily means from start-of-interval to center-of-interval times.
    Only the daily data is currently corrected with an offset of 12 hours.
    This does not shift yearly, monthly, hourly, 3hourly or 6hourly data.
    Also daily averaged datasets called *-d are not corrected.
    https://help.marine.copernicus.eu/en/articles/6820094-how-is-defined-the-nomenclature-of-copernicus-marine-data
    """
    if "P1D-m" in dataset_id or dataset_id.endswith("rean-d"):
        # TODO: remove rean-d https://github.com/Deltares/dfm_tools/issues/1090
        # check if dataset times are indeed at midnight (start-of-interval)
        assert (ds["time"].to_pandas().dt.hour == 0).all()
        # add offset to correct from midnight to noon (center-of-interval)
        print("corrected daily averaged dataset from midnight to noon by "
              "adding a 12-hour offset.")
        ds["time"] = ds["time"] + pd.Timedelta(hours=12)
    return ds


def copernicusmarine_dataset_timerange(dataset_id):
    ds = copernicusmarine.open_dataset(
        dataset_id=dataset_id,
        **COPERNICUSMARINE_OPTIMIZE_ARGS
        )
    ds = copernicusmarine_dataset_timeshift(ds, dataset_id)
    ds_tstart = pd.Timestamp(ds.time.isel(time=0).values)
    ds_tstop = pd.Timestamp(ds.time.isel(time=-1).values)
    return ds_tstart, ds_tstop


def open_OPeNDAP_xr(dataset_url):
    """
    
    How to get the opendap dataset_url (HYCOM example):
        - https://www.hycom.org/dataserver
        - Select a product and search for THREDDS, e.g.: https://www.hycom.org/dataserver/gofs-3pt1/analysis
        - find an opendap dataset_url, it depends per product/run where to find it.
        Some examples:
            https://tds.hycom.org/thredds/dodsC/GLBu0.08/expt_19.1/2010
            https://tds.hycom.org/thredds/dodsC/GLBy0.08/expt_93.0

    """
        
    if isinstance(dataset_url,list):
        dataset_url_one = dataset_url[0]
    else:
        dataset_url_one = dataset_url
    
    if 'hycom.org' in dataset_url_one:
        if isinstance(dataset_url,list):
            print(f'xarray opening opendap dataset like: {dataset_url[0]}.html ({len(dataset_url)} urls/years)')
            data_xr = xr.open_mfdataset(dataset_url,decode_times=False) #TODO: for some reason decode_times does not work: "ValueError: unable to decode time units 'hours since analysis' with 'the default calendar'."
        else:
            print(f'xarray opening opendap dataset: {dataset_url}.html')
            data_xr = xr.open_dataset(dataset_url,decode_times=False) #TODO: for some reason decode_times does not work: "ValueError: unable to decode time units 'hours since analysis' with 'the default calendar
        data_xr['time'] = cftime.num2date(data_xr.time,units=data_xr.time.units,calendar=data_xr.time.calendar)
        data_xr = data_xr.rename({'lon':'longitude','lat':'latitude'})
    else:
        print(f'unspecified dataset_url, might fail: {dataset_url}')
        if isinstance(dataset_url,list):
            print(f'xarray opening opendap dataset like: {dataset_url[0]}.html ({len(dataset_url)} urls/years)')
            data_xr = xr.open_mfdataset(dataset_url)
        else:
            print(f'xarray opening opendap dataset: {dataset_url}.html')
            data_xr = xr.open_dataset(dataset_url)
        if 'lon' in data_xr.dims:
            data_xr = data_xr.rename({'lon':'longitude','lat':'latitude'})
        
    return data_xr


def download_OPeNDAP(dataset_url,
                     varkey,
                     longitude_min, longitude_max, latitude_min, latitude_max, 
                     date_min, date_max, freq='D',
                     dir_output='.', file_prefix='', overwrite=False):
    """
    

    Parameters
    ----------
    dataset_url : TYPE
        DESCRIPTION.
    varkey : TYPE
        DESCRIPTION.
    longitude_min : TYPE
        DESCRIPTION.
    longitude_max : TYPE
        DESCRIPTION.
    latitude_min : TYPE
        DESCRIPTION.
    latitude_max : TYPE
        DESCRIPTION.
    date_min : TYPE
        DESCRIPTION.
    date_max : TYPE
        DESCRIPTION.
    freq : TYPE, optional
        DESCRIPTION. The default is 'D'.
    dir_output : TYPE, optional
        DESCRIPTION. The default is '.'.
    file_prefix : TYPE, optional
        DESCRIPTION. The default is ''.
    overwrite : TYPE, optional
        DESCRIPTION. The default is False.
    
    Raises
    ------
    KeyError
        DESCRIPTION.
    OutOfRangeError
        DESCRIPTION.

    Returns
    -------
    None.

    """
    
    data_xr = open_OPeNDAP_xr(dataset_url=dataset_url)
    
    print(f'xarray subsetting data (variable \'{varkey}\' and lon/lat extents)')
    if varkey not in data_xr.data_vars:
        raise KeyError(f'"{varkey}" not found in dataset, available are: {list(data_xr.data_vars)}')
    data_xr_var = data_xr[[varkey]]
    data_xr_var = data_xr_var.sel(longitude=slice(longitude_min,longitude_max), #TODO: add depth selection?
                                  latitude=slice(latitude_min,latitude_max))
    data_xr_times = data_xr_var.time.to_series()
    print(f'available time range in dataset from {data_xr_times.index[0]} to {data_xr_times.index[-1]}')
    period_range = pd.period_range(date_min,date_max,freq=freq)
    
    #check if date_min/date_max are available in dataset
    if not (data_xr_times.index[0] <= period_range[0].to_timestamp() <= data_xr_times.index[-1]):
        raise OutOfRangeError(f'date_min ({period_range[0]}) is outside available time range in dataset: {data_xr_times.index[0]} to {data_xr_times.index[-1]}')
    if not (data_xr_times.index[0] <= period_range[-1].to_timestamp() <= data_xr_times.index[-1]):
        raise OutOfRangeError(f'date_max ({period_range[-1]}) is outside available time range in dataset: {data_xr_times.index[0]} to {data_xr_times.index[-1]}')
    
    for date in period_range:
        date_str = str(date)
        name_output = f'{file_prefix}{varkey}_{date_str}.nc'
        file_out = Path(dir_output,name_output)
        if file_out.is_file() and not overwrite:
            print(f'"{name_output}" found and overwrite=False, continuing.')
            continue
        
        print(f'xarray subsetting data per {period_range.freq}: {date_str}')
        data_xr_var_seltime = data_xr_var.sel(time=slice(date_str,date_str)) #+' 12:00:00', 
        
        print(f'xarray writing netcdf file: {name_output}')
        data_xr_var_seltime.to_netcdf(os.path.join(dir_output,name_output)) #TODO: add chunks={'time':1} or only possible with opening?
        data_xr_var_seltime.close()



import os
import numpy as np
import xugrid as xu
import xarray as xr
import datetime as dt
import pandas as pd
import meshkernel
from dfm_tools.xarray_helpers import file_to_list
from netCDF4 import default_fillvals
import warnings

__all__ = [
    "open_partitioned_dataset",
    "open_dataset_curvilinear",
    "open_dataset_delft3d4",
    "uda_to_faces",
    "uda_interfaces_to_centers",
    "add_network_cellinfo",
    "enrich_rst_with_map",
]


def get_vertical_dimensions(uds): #TODO: maybe add layer_dimension and interface_dimension properties to xugrid?
    """
    get vertical_dimensions from grid_info of ugrid mapfile (this will fail for hisfiles). The info is stored in the layer_dimension and interface_dimension attribute of the mesh2d variable of the dataset (stored in uds.grid after reading with xugrid)
    
    processing cb_3d_map.nc
        >> found layer/interface dimensions in file: mesh2d_nLayers mesh2d_nInterfaces
    processing Grevelingen-FM_0*_map.nc
        >> found layer/interface dimensions in file: nmesh2d_layer nmesh2d_interface (these are updated in open_partitioned_dataset)
    processing DCSM-FM_0_5nm_0*_map.nc
        >> found layer/interface dimensions in file: mesh2d_nLayers mesh2d_nInterfaces
    processing MB_02_0*_map.nc
        >> found layer/interface dimensions in file: mesh2d_nLayers mesh2d_nInterfaces
    """
        
    if not hasattr(uds,'grid'): #early return in case of e.g. hisfile
        return None, None
        
    gridname = uds.grid.name
    grid_info = uds.grid.to_dataset()[gridname]
    if hasattr(grid_info,'layer_dimension'):
        return grid_info.layer_dimension, grid_info.interface_dimension
    else: # 2D model or networkfile
        return None, None


def remove_ghostcells(uds, fname): #TODO: remove ghostcells from output or align values between (non)ghost cells: https://issuetracker.deltares.nl/browse/UNST-6701
    """
    Dropping ghostcells if there is a domainno variable present and there is a domainno in the filename.
    Not using most-occurring domainno in var, since this is not a valid assumption for merged datasets and might be invalid for a very small partition.
    
    """
    gridname = uds.grid.name
    varn_domain = f'{gridname}_flowelem_domain'
    
    #check if dataset has domainno variable, return uds if not present
    if varn_domain not in uds.data_vars:
        print('[nodomainvar] ',end='')
        return uds
    
    #derive domainno from filename, return uds if not present
    if '_' not in fname: #safety escape in case there is no _ in the filename
        print('[nodomainfname] ',end='')
        return uds
    fname_splitted = fname.split('_')
    part_domainno_fromfname = fname_splitted[-2] #this is not valid for rstfiles (date follows after partnumber), but they cannot be read with xugrid anyway since they lack topology and node_x/node_y variables: https://issuetracker.deltares.nl/browse/UNST-7176
    if not part_domainno_fromfname.isnumeric() or len(part_domainno_fromfname)!=4:
        print('[nodomainfname] ',end='')
        return uds
    
    #drop ghostcells
    part_domainno_fromfname = int(part_domainno_fromfname)
    da_domainno = uds.variables[varn_domain]
    idx = np.flatnonzero(da_domainno == part_domainno_fromfname)
    uds = uds.isel({uds.grid.face_dimension:idx})
    return uds


def remove_unassociated_edges(ds: xr.Dataset) -> xr.Dataset:
    """
    Removes edges that are not associated to any of the faces, usecase in https://github.com/Deltares/xugrid/issues/68

    Parameters
    ----------
    ds : xr.Dataset
        DESCRIPTION.
    topology : str, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    ds : xr.Dataset
        DESCRIPTION.

    """
    
    uds = xu.core.wrap.UgridDataset(ds)
    
    # escape for 1D networks
    if isinstance(uds.grid,xu.Ugrid1d):
        return ds
    
    associated = uds.grid.validate_edge_node_connectivity()
    edge_dimension = uds.grid.edge_dimension
    
    if not associated.all():
         print(f"[{(~associated).sum()} unassociated edges removed] ",end='')
         ds = ds.sel({edge_dimension: associated})
    return ds


def decode_default_fillvals(ds):
    """
    xarray only supports explicitly set _FillValue attrs, and therefore ignores the default netCDF4 fillvalue
    This function adds the default fillvalue as _FillValue attribute and decodes the dataset again.
    """
    # TODO: this function can be removed when xarray does it automatically: https://github.com/Deltares/dfm_tools/issues/490
    
    nfillattrs_added = 0
    for varn in ds.variables.keys():
        # TODO: possible to get always_mask boolean with `netCDF4.Dataset(file_nc).variables[varn].always_mask`, but this seems to be always True for FM mapfiles
        if '_FillValue' in ds.variables[varn].encoding:
            continue
        dtype_str = ds.variables[varn].dtype.str[1:]
        if dtype_str not in default_fillvals.keys():
            continue
        varn_fillval = default_fillvals[dtype_str]
        ds[varn] = ds[varn].assign_attrs({'_FillValue':varn_fillval})
        nfillattrs_added += 1
    print(f'[default_fillvals decoded for {nfillattrs_added} variables] ',end='')
    
    #decode the dataset with newly added _FillValue attrs again
    ds = xr.decode_cf(ds)
    return ds


def remove_nan_fillvalue_attrs(ds : (xr.Dataset, xu.UgridDataset)):
    """
    xarray writes {"_FillValue": np.nan} to encoding for variables without _FillValue attribute.
    Remove these again upon reading to avoid issues.
    """
    if isinstance(ds,xu.UgridDataset):
        ds = ds.obj
    
    count = 0
    for varn in ds.variables.keys():
        if '_FillValue' in ds.variables[varn].encoding:
            if np.isnan(ds.variables[varn].encoding['_FillValue']):
                ds[varn].encoding.pop('_FillValue')
                count += 1
    if count > 0:
        print(f"[{count} nan fillvalue attrs removed]", end="")


def uds_auto_set_crs(uds : xu.UgridDataset):
    # FM-mapfiles contain wgs84/projected_coordinate_system variables with epsg attr, xugrid has .crs property
    # TODO: parse+set crs in xugrid instead: https://github.com/Deltares/xugrid/issues/42
    # also adjusting projected_coordinate_system/wgs84 when using set_crs/to_crs
    
    uds_epsg = uds.filter_by_attrs(epsg=lambda v: v is not None)
    if len(uds_epsg.data_vars) != 1:
        return
    
    crs_varn = list(uds_epsg.data_vars)[0]
    epsg = uds[crs_varn].attrs["epsg"]
    from pyproj.exceptions import CRSError
    try:
        uds.ugrid.set_crs(epsg)
    except CRSError:
        return


def open_partitioned_dataset(file_nc:str, decode_fillvals:bool = False, remove_edges:bool = False, remove_ghost:bool = True, **kwargs): 
    """
    using xugrid to read and merge partitions, including support for delft3dfm mapformat1 
    by renaming old layerdim. Furthermore some optional extensions like removal of hanging
    edges and removal of ghost cells.

    Parameters
    ----------
    file_nc : str
        DESCRIPTION.
    decode_fillvals : bool, optional
        DESCRIPTION. The default is False.
    remove_edges : bool, optional
        Remove hanging edges from the mapfile, necessary to generate contour and contourf plots 
        with xugrid. Can be enabled always, but takes some additional computation time. The default is False.
    remove_ghost : bool, optional
        Remove ghostcells from the partitions. This is also done by xugrid automatically 
        upon merging, but then the domain numbers are not taken into account so 
        the result will be different. The default is True.
    file_nc : TYPE
        DESCRIPTION.
    kwargs : TYPE, optional
        arguments that are passed to xr.open_dataset. The chunks argument is set if not provided
        chunks={'time':1} increases performance significantly upon reading, but causes memory overloads when performing sum/mean/etc actions over time dimension (in that case 100/200 is better). The default is {'time':1}.

    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    ds_merged_xu : TYPE
        DESCRIPTION.
    
    """
    # TODO: timings
    # file_nc = 'p:\\1204257-dcsmzuno\\2006-2012\\3D-DCSM-FM\\A18b_ntsu1\\DFM_OUTPUT_DCSM-FM_0_5nm\\DCSM-FM_0_5nm_0*_map.nc' #3D DCSM
    # file_nc = 'p:\\archivedprojects\\11206813-006-kpp2021_rmm-2d\\C_Work\\31_RMM_FMmodel\\computations\\model_setup\\run_207\\results\\RMM_dflowfm_0*_map.nc' #RMM 2D
    # file_nc = 'p:\\1230882-emodnet_hrsm\\GTSMv5.0\\runs\\reference_GTSMv4.1_wiCA_2.20.06_mapformat4\\output\\gtsm_model_0*_map.nc' #GTSM 2D
    # file_nc = 'p:\\11208053-005-kpp2022-rmm3d\\C_Work\\01_saltiMarlein\\RMM_2019_computations_02\\computations\\theo_03\\DFM_OUTPUT_RMM_dflowfm_2019\\RMM_dflowfm_2019_0*_map.nc' #RMM 3D
    # file_nc = 'p:\\archivedprojects\\11203379-005-mwra-updated-bem\\03_model\\02_final\\A72_ntsu0_kzlb2\\DFM_OUTPUT_MB_02\\MB_02_0*_map.nc'
    # Timings (xu.open_dataset/xu.merge_partitions):
    # - DCSM 3D 20 partitions  367 timesteps: 231.5/ 4.5 sec (decode_times=False: 229.0 sec)
    # - RMM  2D  8 partitions  421 timesteps:  55.4/ 4.4 sec (decode_times=False:  56.6 sec)
    # - GTSM 2D  8 partitions  746 timesteps:  71.8/30.0 sec (decode_times=False: 204.8 sec)
    # - RMM  3D 40 partitions  146 timesteps: 168.8/ 6.3 sec (decode_times=False: 158.4 sec)
    # - MWRA 3D 20 partitions 2551 timesteps:  74.4/ 3.4 sec (decode_times=False:  79.0 sec)

    #TODO: add support for multiple grids via keyword? https://github.com/Deltares/dfm_tools/issues/497
    #TODO: speed up open_dataset https://github.com/Deltares/dfm_tools/issues/225 (also remove_ghost)
    
    if 'chunks' not in kwargs:
        kwargs['chunks'] = {'time':1}
    if 'decode_timedelta' not in kwargs:
        # avoid futurewarning: https://github.com/Deltares/dfm_tools/issues/1100
        kwargs['decode_timedelta'] = False
    
    dtstart_all = dt.datetime.now()
    file_nc_list = file_to_list(file_nc)
    
    print(f'>> xu.open_dataset() with {len(file_nc_list)} partition(s): ',end='')
    dtstart = dt.datetime.now()
    partitions = []
    for iF, file_nc_one in enumerate(file_nc_list):
        print(iF+1,end=' ')
        # suppress chunking warning: https://github.com/Deltares/dfm_tools/issues/947
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            ds = xr.open_mfdataset(file_nc_one, **kwargs)
        if decode_fillvals:
            ds = decode_default_fillvals(ds)
        if remove_edges:
            ds = remove_unassociated_edges(ds)
        if 'nFlowElem' in ds.dims and 'nNetElem' in ds.dims:
            print('[mapformat1] ',end='')
            #for mapformat1 mapfiles: merge different face dimensions (rename nFlowElem to nNetElem) to make sure the dataset topology is correct
            ds = ds.rename({'nFlowElem':'nNetElem'})
        remove_nan_fillvalue_attrs(ds)
        uds = xu.core.wrap.UgridDataset(ds)
        if remove_ghost: #TODO: this makes it way slower (at least for GTSM, although merging seems faster), but is necessary since values on overlapping cells are not always identical (eg in case of Venice ucmag)
            uds = remove_ghostcells(uds, file_nc_one)
        uds_auto_set_crs(uds)
        partitions.append(uds)
    print(': ',end='')
    print(f'{(dt.datetime.now()-dtstart).total_seconds():.2f} sec')
    
    if len(partitions) == 1: #do not merge in case of 1 partition
        return partitions[0]
    
    print(f'>> xu.merge_partitions() with {len(file_nc_list)} partition(s): ',end='')
    dtstart = dt.datetime.now()
    ds_merged_xu = xu.merge_partitions(partitions)
    print(f'{(dt.datetime.now()-dtstart).total_seconds():.2f} sec')
    
    #print variables that are dropped in merging procedure. Often only ['mesh2d_face_x_bnd', 'mesh2d_face_y_bnd'], which can be derived by combining node_coordinates (mesh2d_node_x mesh2d_node_y) and face_node_connectivity (mesh2d_face_nodes). >> can be removed from FM-mapfiles (email of 16-1-2023)
    varlist_onepart = list(partitions[0].variables.keys())
    varlist_merged = list(ds_merged_xu.variables.keys())
    varlist_dropped_bool = ~pd.Series(varlist_onepart).isin(varlist_merged)
    varlist_dropped = pd.Series(varlist_onepart).loc[varlist_dropped_bool]
    if varlist_dropped_bool.any():
        print(f'>> some variables dropped with merging of partitions: {varlist_dropped.tolist()}')
    
    print(f'>> dfmt.open_partitioned_dataset() total: {(dt.datetime.now()-dtstart_all).total_seconds():.2f} sec')
    return ds_merged_xu


def open_dataset_curvilinear(file_nc,
                             x_dim:str,
                             y_dim:str,
                             x_bounds:str,
                             y_bounds:str,
                             convert_360to180:bool = False ,
                             **kwargs) -> xu.UgridDataset:
    """
    Construct a UgridDataset from a curvilinear grid with 2D lat/lon variables
    with i/j indexes/dims, including vertices variables. Works for curvilinear
    datasets like CMCC and also for WAQUA files that are converted with getdata

    Parameters
    ----------
    file_nc : str or path
        DESCRIPTION.
    x_dim : str
        The x-dimension, like lon, i or N.
    y_dim : str
        The y-dimension, like lat, j or M.
    x_bounds : str
        The variable with the x-bounds, like vertices_longitude or grid_x.
    y_bounds : str
        The variable with the y-bounds, like vertices_latitude or grid_y.
    convert_360to180 : bool, optional
        Whether to convert from a 0 to 360 degree global model to a -180 to 180
        degree global model. The default is False.
    **kwargs : TYPE
        additional arguments are passed on to xr.open_mfdataset().

    Returns
    -------
    uds : xu.UgridDataset
        The resulting ugrid dataset.

    """
    
    if 'chunks' not in kwargs:
        kwargs['chunks'] = {'time':1}
    
    # data_vars='minimal' to avoid time dimension on vertices_latitude and
    # others when opening multiple files at once
    print('>> open_mfdataset: ',end='')
    dtstart = dt.datetime.now()
    ds = xr.open_mfdataset(file_nc, data_vars="minimal", **kwargs)
    print(f'{(dt.datetime.now()-dtstart).total_seconds():.2f} sec')

    # convert from 0to360 to -180 to 180
    if convert_360to180:
        ds[x_bounds] = (ds[x_bounds]+180) % 360 - 180
    
    topology = {"mesh2d":{"x":x_dim,
                          "y":y_dim,
                          "x_bounds":x_bounds,
                          "y_bounds":y_bounds,
                          },
                }
    
    print('>> convert to xugrid.UgridDataset: ',end='')
    dtstart = dt.datetime.now()
    uds = xu.UgridDataset.from_structured2d(ds, topology=topology)
    print(f'{(dt.datetime.now()-dtstart).total_seconds():.2f} sec')
    return uds


def delft3d4_get_nanmask(x,y):
    # -999.999 in kivu and 0.0 in curvedbend, both in westernscheldt
    bool_0 = (x==0) & (y==0)
    bool_1 = (x==-999) & (y==-999)
    bool_2 = (x==-999.999) & (y==-999.999)
    bool_mask = bool_0 | bool_1 | bool_2
    return bool_mask


def delft3d4_stack_shifted_coords(da):
    shift = 1
    np_stacked = np.stack([
        da, #ll
        da.shift(MC=shift), #lr
        da.shift(MC=shift, NC=shift), #ur
        da.shift(NC=shift), #ul
        ],axis=-1)
    da_stacked = xr.DataArray(np_stacked, dims=("M","N","four"))
    return da_stacked


def delft3d4_convert_uv(ds):
    # replace invalid values not with nan but with zero
    # otherwise the spatial coverage is affected
    mask_u1 = (ds.U1==-999) | (ds.U1==-999.999)
    mask_v1 = (ds.V1==-999) | (ds.V1==-999.999)
    u1_mn = ds.U1.where(~mask_u1, 0)
    v1_mn = ds.V1.where(~mask_v1, 0)
    
    # minus 0.5 since padding=low so corner value is representative for
    # previous face according to that logic, method=nearest might be better
    # (or just rename the dims)
    u1_mn_cen = u1_mn.interp(MC=u1_mn.MC-0.5, method='linear')
    v1_mn_cen = v1_mn.interp(NC=v1_mn.NC-0.5, method='linear')
    # rename corner dims to center dims since we shifted them with 0.5
    u1_mn_cen = u1_mn_cen.rename({'MC':'M'})
    v1_mn_cen = v1_mn_cen.rename({'NC':'N'})
    # TODO: since padding=low, just renaming the dims might even be better?
    # u1_mn_cen = u1_mn.rename({'MC':'M'})
    # v1_mn_cen = v1_mn.rename({'NC':'N'})
    # >> could also be done with `ds = ds.swap_dims({"M":"MC","N":"NC"})`
    
    # create combined uv mask (have to rename dimensions)
    mask_u1_mn = mask_u1.rename({'MC':'M'})
    mask_v1_mn = mask_v1.rename({'NC':'N'})
    mask_uv1_mn = mask_u1_mn & mask_v1_mn
    # drop all actual missing cells
    u1_mn_cen = u1_mn_cen.where(~mask_uv1_mn)
    v1_mn_cen = v1_mn_cen.where(~mask_uv1_mn)
    
    # to avoid creating large chunks, alternative is to overwrite the vars
    # with the MN-averaged vars, but it requires passing and updating of attrs
    ds = ds.drop_vars(['U1','V1'])
    
    # compute ux/uy/umag/udir
    # TODO: add attrs to variables
    alfas_rad = np.deg2rad(ds.ALFAS)
    vel_x = u1_mn_cen*np.cos(alfas_rad) - v1_mn_cen*np.sin(alfas_rad)
    vel_y = u1_mn_cen*np.sin(alfas_rad) + v1_mn_cen*np.cos(alfas_rad)
    ds['ux'] = vel_x
    ds['uy'] = vel_y
    ds['umag'] = np.sqrt(vel_x**2 + vel_y**2)
    ds['udir'] = np.rad2deg(np.arctan2(vel_y, vel_x))%360
    return ds


def open_dataset_delft3d4(file_nc, **kwargs) -> xu.UgridDataset:
    """
    Reads in a Delft3D4 netcdf outputfile (curvilinear/staggered) as a ugrid
    (xugrid.UgridDataset) dataset. This is a Delft3D4 specific version of
    dfmt.open_dataset_curvilinear().
    
    To get Delft3D4 to write netCDF output instead of .dat files, add these
    lines to your model settings file (.mdf):
    
    - FlNcdf=#maphis#
    - ncFormat=4

    Parameters
    ----------
    file_nc : str or path
        DESCRIPTION.
    **kwargs : TYPE
        additional arguments are passed on to xr.open_mfdataset().

    Returns
    -------
    uds : xu.UgridDataset
        The resulting ugrid dataset.

    """
    
    if 'chunks' not in kwargs:
        kwargs['chunks'] = {'time':1}
    
    ds = xr.open_dataset(file_nc, **kwargs)
    
    # prevent grid variable that might be confused with uds.grid accessor
    if 'grid' in ds.data_vars:
        ds = ds.rename_vars({'grid':'grid_original'})
    
    xcor_stacked = delft3d4_stack_shifted_coords(ds.XCOR)
    ycor_stacked = delft3d4_stack_shifted_coords(ds.YCOR)
    mask_xy = delft3d4_get_nanmask(xcor_stacked,ycor_stacked)
    ds['xcor_stacked'] = xcor_stacked.where(~mask_xy)
    ds['ycor_stacked'] = ycor_stacked.where(~mask_xy)
    
    if ('U1' in ds.data_vars) and ('V1' in ds.data_vars):
        ds = delft3d4_convert_uv(ds)
    
    # TODO: consider using same dims for variables on cell corners and faces
    # ds = ds.swap_dims({"M":"MC","N":"NC"})
    topology = {"mesh2d":{"x":"M",
                          "y":"N",
                          "x_bounds":"xcor_stacked",
                          "y_bounds":"ycor_stacked",
                          }
                }
    uds = xu.UgridDataset.from_structured2d(ds, topology=topology)
    
    return uds


def uda_to_faces(uda : xu.UgridDataArray) -> xu.UgridDataArray:
    """
    Interpolates a ugrid variable (xu.DataArray) with a node or edge dimension
    to the faces by averaging the 3/4 nodes/edges around each face.
    
    Parameters
    ----------
    uda : xu.UgridDataArray
        DESCRIPTION.


    Returns
    -------
    uda_face : xu.UgridDataArray
        DESCRIPTION.

    """
    grid = uda.grid
    
    dimn_faces = grid.face_dimension
    
    # construct indexing array
    if dimn_faces in uda.dims:
        print(f'provided uda/variable "{uda.name}" does not have an node or edge dimension, returning unchanged uda')
        return uda
    
    # couple to faces and take mean
    print('to_face() interpolation: ',end='')
    dtstart = dt.datetime.now()
    uda_face = uda.ugrid.to_face().mean(dim="nmax", keep_attrs=True)
    print(f'{(dt.datetime.now()-dtstart).total_seconds():.2f} sec')

    #update attrs from node/edge to face
    face_attrs = {'location': 'face', 'cell_methods': f'{dimn_faces}: mean'}
    uda_face = uda_face.assign_attrs(face_attrs)
    
    return uda_face


def uda_interfaces_to_centers(uda_int : xu.UgridDataArray) -> xu.UgridDataArray:
    dimn_layer, dimn_interface = get_vertical_dimensions(uda_int)
    
    if dimn_interface not in uda_int.dims:
        print('no interface dimension found, returning original array')
        return uda_int
        
    # define layers to be halfway inbetween the interfaces
    nlayers = uda_int.sizes[dimn_interface] - 1
    array_shift_half = xr.DataArray(np.arange(0.5,nlayers),dims=dimn_layer)
    # interpolate from interfaces to layers
    uda_cen = uda_int.interp({dimn_interface:array_shift_half},assume_sorted=True)
    # drop interface coordinate
    uda_cen = uda_cen.drop_vars(dimn_interface)
    
    return uda_cen


def add_network_cellinfo(uds:xu.UgridDataset):
    """
    Reads a UgridDataset with a minimal topology as it occurs in dflowfm netfiles,
    this contains only node_x, node_y and edge_node_connectivity. xugrid reads this as Ugrid1d
    It converts it to a Ugrid2d grid which includes the face_node_connectivity.
    It also couples the variables like NetNode_z again, but drops the edge_dims to avoid conflicts.
    """
    #check if indeed 1D grid object
    assert isinstance(uds.grid, xu.Ugrid1d)
    
    # derive meshkernel from grid with Mesh1d
    mk1 = uds.grid.meshkernel
    mk_mesh1d = mk1.mesh1d_get()
    crs = uds.grid.crs
    
    # use Mesh1d nodes/edgenodes info for generation of meshkernel with Mesh2d
    is_geographic = uds.grid.is_geographic
    from dfm_tools.meshkernel_helpers import geographic_to_meshkernel_projection
    projection = geographic_to_meshkernel_projection(is_geographic)
    mk_mesh2d = meshkernel.Mesh2d(mk_mesh1d.node_x, mk_mesh1d.node_y, mk_mesh1d.edge_nodes)
    mk2 = meshkernel.MeshKernel(projection=projection)
    mk2.mesh2d_set(mk_mesh2d)
    mesh2d_grid = mk2.mesh2d_get() #this is a more populated version of mk_mesh2d, needed for xugrid
    #TODO: we have to supply is_geographic twice, necessary?
    # also "projected" is opposite of "is_geographic" according to the docstring
    xu_grid = xu.Ugrid2d.from_meshkernel(mesh2d_grid, projected = not is_geographic, crs=crs)
    
    # convert uds.obj (non-grid vars from dataset) to new xugrid standards
    rename_dims_dict = {uds.grid.node_dimension:xu_grid.node_dimension,
                        # uds.grid.edge_dimension:xu_grid.edge_dimension,
                        }
    uds_obj = uds.obj.rename_dims(rename_dims_dict)
    # drop edge dim since dim size can be changed in case of hanging edges
    uds_obj = uds_obj.drop_dims(uds.grid.edge_dimension)
    # drop node coordinate var since the order might conflict with the new grid and it has no meaning
    uds_obj = uds_obj.drop_vars(uds.grid.node_dimension)
    
    # combine again with rest of dataset
    uds_withinfo = xu.UgridDataset(obj=uds_obj, grids=[xu_grid])
    
    return uds_withinfo


def enrich_rst_with_map(ds_rst:xr.Dataset):
    """
    enriches rst dataset with topology from mapfile in the same folder.
    in order to merge, the bnd dimension has to be dropped.
    Also, the domain variable is currently not merged, so ghostcells are not removed.

    Parameters
    ----------
    ds_rst : xr.Dataset
        DESCRIPTION.

    Returns
    -------
    ds_rst : TYPE
        DESCRIPTION.

    """
    
    # derive fname of mapfile and open dataset
    file_nc_rst = ds_rst.encoding["source"]
    dir_output = os.path.dirname(file_nc_rst)
    fname_rst = os.path.basename(file_nc_rst)
    fname_map = "_".join(fname_rst.split("_")[:-3]) + "_map.nc"
    file_nc_map = os.path.join(dir_output, fname_map)
    ds_map = xr.open_dataset(file_nc_map)
    
    # remove unassociated edges from mapfile to align with rst file
    ds_map = remove_unassociated_edges(ds_map)
    
    # enrich rst file with topology variables from mapfile
    topology_varn = ds_map.ugrid_roles.topology[0]
    topo_var = ds_map[topology_varn]
    ds_rst[topology_varn] = topo_var
    nodecoords = topo_var.attrs["node_coordinates"]
    for topovar in nodecoords.split():
        ds_rst[topovar] = ds_map[topovar]
        ds_rst = ds_rst.set_coords(topovar)
    edgecoords = topo_var.attrs["edge_coordinates"]
    for topovar in edgecoords.split():
        ds_rst[topovar] = ds_map[topovar]
        ds_rst = ds_rst.set_coords(topovar)
    fnc_varn = topo_var.attrs["face_node_connectivity"]
    ds_rst[fnc_varn] = ds_map[fnc_varn]
    
    # rename old dims
    if 'nFlowLinkPts' in ds_rst.dims and 'nNetLinkPts' in ds_rst.dims:
        ds_rst = ds_rst.rename({'nFlowLinkPts':'nNetLinkPts'})
    if 'nFlowElem' in ds_rst.dims:
        facedim = topo_var.attrs["face_dimension"]
        ds_rst = ds_rst.rename({'nFlowElem':facedim})
    if 'nNetElem' in ds_rst.dims:
        facedim = topo_var.attrs["face_dimension"]
        ds_rst = ds_rst.rename({'nNetElem':facedim})
    if "nNetLink" in ds_rst.dims:
        edgedim = topo_var.attrs["edge_dimension"]
        ds_rst = ds_rst.rename({"nNetLink":edgedim})
    if "nNetElemMaxNode" in ds_rst.dims:
        mfndim = topo_var.attrs["max_face_nodes_dimension"]
        ds_rst = ds_rst.rename({"nNetElemMaxNode":mfndim})
    
    # drop bnd dim since it cannot be merged by xugrid
    if "nFlowElemBnd" in ds_rst.dims:
        ds_rst = ds_rst.drop_dims("nFlowElemBnd")
    
    return ds_rst



# -*- coding: utf-8 -*-
"""
Created on Fri Oct 20 16:09:20 2023

@author: veenstra
"""

import os
import pytest
import numpy as np
import pandas as pd
import geopandas as gpd
import dfm_tools as dfmt
import hydrolib.core.dflowfm as hcdfm
from dfm_tools.hydrolib_helpers import get_ncbnd_construct


@pytest.mark.unittest
def test_geodataframe_with_Polygon_to_PolyFile():
    """
    converting a geodataframe with Polygon geometries to hcdfm.PolyFile
    """
    lon_min, lon_max, lat_min, lat_max = -6, 2, 48.5, 51.2
    bbox = (lon_min, lat_min, lon_max, lat_max)
    
    # gdf containing shapely.Polygon geometries
    coastlines_gdf = dfmt.get_coastlines_gdb(bbox=bbox, res='h')
    
    polyfile = dfmt.geodataframe_to_PolyFile(coastlines_gdf)
    assert isinstance(polyfile, hcdfm.PolyFile)


@pytest.mark.unittest
def test_geodataframe_with_LineString_to_PolyFile(tmp_path):
    """
    converting a geodataframe with Linestring geometries to hcdfm.PolyFile
    """
    # write polygon file
    file_pol = tmp_path / 'temp_coastlines.pol'
    lon_min, lon_max, lat_min, lat_max = -68.45, -68.1, 12, 12.35
    bbox = (lon_min, lat_min, lon_max, lat_max)
    coastlines_gdf = dfmt.get_coastlines_gdb(bbox=bbox, res='h')
    polyfile = dfmt.geodataframe_to_PolyFile(coastlines_gdf)
    polyfile.save(file_pol)
    
    # read polygon file
    polyfile_object = hcdfm.PolyFile(file_pol)

    # gdf containing shapely.LineString geometries
    gdf_polyfile = dfmt.PolyFile_to_geodataframe_linestrings(polyfile_object,crs=None)
    
    polyfile = dfmt.geodataframe_to_PolyFile(gdf_polyfile)
    assert isinstance(polyfile, hcdfm.PolyFile)


@pytest.fixture(scope='session')
def bnd_gdf():
    dxy = 0.02
    crs = 4326
    lon_min, lon_max, lat_min, lat_max = -68.31, -68.27, 12.10, 12.21
    mk_object = dfmt.make_basegrid(lon_min, lon_max, lat_min, lat_max, dx=dxy, dy=dxy, crs=crs)
    bnd_gdf = dfmt.generate_bndpli_cutland(mk=mk_object, res='h', buffer=0.01)
    return bnd_gdf


@pytest.mark.unittest
def test_geodataframe_to_PolyFile_name_default(bnd_gdf):
    polyfile_obj = dfmt.geodataframe_to_PolyFile(bnd_gdf)
    names = [x.metadata.name for x in polyfile_obj.objects]
    assert names == ['L1', 'L2']


@pytest.mark.unittest
def test_geodataframe_to_PolyFile_name_some(bnd_gdf):
    polyfile_obj = dfmt.geodataframe_to_PolyFile(bnd_gdf, name="test_model")
    names = [x.metadata.name for x in polyfile_obj.objects]
    assert names == ['test_model1', 'test_model2']


@pytest.mark.unittest
def test_geodataframe_to_PolyFile_name_invalidtype(bnd_gdf):
    with pytest.raises(TypeError) as e:
        dfmt.geodataframe_to_PolyFile(bnd_gdf, name=None)
    assert 'name should be a string' in str(e.value)


@pytest.mark.unittest
def test_geodataframe_to_PolyFile_name_incorrect(bnd_gdf):
    with pytest.raises(ValueError) as e:
        dfmt.geodataframe_to_PolyFile(bnd_gdf, name='1')
    assert 'names in polyfile do not all start with a letter' in str(e.value)
    
    with pytest.raises(ValueError) as e:
        dfmt.geodataframe_to_PolyFile(bnd_gdf, name='-')
    assert 'names in polyfile do not all start with a letter' in str(e.value)
    
    with pytest.raises(ValueError) as e:
        dfmt.geodataframe_to_PolyFile(bnd_gdf, name='')
    assert 'names in polyfile do not all start with a letter' in str(e.value)


@pytest.mark.unittest
def test_geodataframe_to_PolyFile_namecolumn_some(bnd_gdf):
    bnd_gdf['name'] = ['test_model1','test_model2']
    polyfile_obj = dfmt.geodataframe_to_PolyFile(bnd_gdf)
    names = [x.metadata.name for x in polyfile_obj.objects]
    assert names == ['test_model1', 'test_model2']


@pytest.mark.unittest
def test_geodataframe_to_PolyFile_namecolumn_name_both(bnd_gdf):
    # name argument is ignored if name column is provided
    # not per se desired, but also not completely wrong
    bnd_gdf['name'] = ['test_model1','test_model2']
    polyfile_obj = dfmt.geodataframe_to_PolyFile(bnd_gdf, name='dummy')
    names = [x.metadata.name for x in polyfile_obj.objects]
    assert names == ['test_model1', 'test_model2']


@pytest.mark.unittest
def test_geodataframe_to_PolyFile_namecolumn_duplicated_names(bnd_gdf):
    # deliberately giving all polylines the same name
    bnd_gdf['name'] = 'duplicate_bnd'
    with pytest.raises(ValueError) as e:
        dfmt.geodataframe_to_PolyFile(bnd_gdf)
    assert 'duplicate polyline names found in polyfile' in str(e.value)

    # deliberately giving all polylines the same empty name
    bnd_gdf['name'] = ''
    with pytest.raises(ValueError) as e:
        dfmt.geodataframe_to_PolyFile(bnd_gdf)
    assert "duplicate polyline names found in polyfile" in str(e.value)


@pytest.mark.unittest
def test_geodataframe_to_PolyFile_namecolumn_numeric_start(bnd_gdf):
    bnd_gdf['name'] = ['1','2']
    with pytest.raises(ValueError) as e:
        dfmt.geodataframe_to_PolyFile(bnd_gdf)
    assert 'names in polyfile do not all start with a letter' in str(e.value)


@pytest.mark.unittest
def test_read_polyfile_as_gdf_points_linestrings(tmp_path):
    ncbnd_construct = get_ncbnd_construct()
    varn_pointname = ncbnd_construct['varn_pointname']
    
    # write a dummy polyfile with first three points of reference file
    # p:\archivedprojects\11208054-004-dcsm-fm\models\model_input\bnd_cond\pli\DCSM-FM_OB_all_20181108.pli
    poly_str = """DCSM-FM_OB_all_20181108
    3    2
    -9.25  43.0
    -9.50  43.0
    -9.75  43.0
    """
    file_pli = os.path.join(tmp_path, "dummy.pli")
    with open(file_pli, "w") as f:
        f.write(poly_str)
    
    # read polyfile with hydrolib-core
    polyfile_object = hcdfm.PolyFile(file_pli)
    
    # convert to geodataframe with points
    gdf_points = dfmt.PolyFile_to_geodataframe_points(polyfile_object)
    
    reference_names = ['DCSM-FM_OB_all_20181108_0001', 'DCSM-FM_OB_all_20181108_0002', 'DCSM-FM_OB_all_20181108_0003']
    reference_x = np.array([-9.25, -9.5 , -9.75])
    reference_y = np.array([43., 43., 43.])
    
    assert isinstance(gdf_points, gpd.GeoDataFrame)
    assert len(gdf_points) == 3
    assert (gdf_points.geometry.type == "Point").all()
    assert np.allclose(gdf_points.geometry.x.to_numpy(), reference_x)
    assert np.allclose(gdf_points.geometry.y.to_numpy(), reference_y)
    assert varn_pointname in gdf_points.columns
    assert gdf_points[varn_pointname].tolist() == reference_names

    # convert to geodataframe with linestrings
    gdf_lines = dfmt.PolyFile_to_geodataframe_linestrings(polyfile_object)
    line0_geom = gdf_lines.geometry.iloc[0]
    
    reference_names = ['DCSM-FM_OB_all_20181108']
    reference_x = np.array([-9.25, -9.5 , -9.75])
    reference_y = np.array([43., 43., 43.])
    
    assert isinstance(gdf_lines, gpd.GeoDataFrame)
    assert len(gdf_lines) == 1
    assert (gdf_lines.geometry.type == "LineString").all()
    assert np.allclose(line0_geom.xy[0], reference_x)
    assert np.allclose(line0_geom.xy[1], reference_y)
    assert 'name' in gdf_lines.columns
    assert gdf_lines['name'].tolist() == reference_names


@pytest.mark.unittest
def test_timmodel_to_dataframe(tmp_path):
    # original data from p:\archivedprojects\11206811-002-d-hydro-grevelingen\simulaties\Jaarsom2017_dfm_006_zlayer\boundary_conditions\hist\jaarsom_2017\sources_sinks\FlakkeeseSpuisluis.tim
    file_tim = os.path.join(tmp_path, "FlakkeeseSpuisluis_dummy.tim")
    str_tim = """* COLUMN_N=25 
     * COLUMN1=Time (min) since the reference date (2016-01-01)
     * COLUMN2=Flow discharge (m3/s), positive in
     * COLUMN3=Salinity (psu)
     * COLUMN4=Temperature (deg)
    * Time(min)   Flow         Salinity       Temperature
    0.00000e+00   0.0000e+00   2.883150e+01   8.4000e+00
    6.56639e+05   0.0000e+00   2.826540e+01   1.0295e+01
    6.56640e+05  -2.1468e+02   2.825750e+01   1.0300e+01
    6.56650e+05  -2.1384e+02   2.808100e+01   1.0400e+01
    6.56660e+05  -2.1384e+02   2.814450e+01   1.0350e+01
    6.56670e+05  -2.1128e+02   2.789250e+01   1.0300e+01
    6.56680e+05  -2.0955e+02   2.788500e+01   1.0350e+01
    6.56690e+05  -2.0781e+02   2.791250e+01   1.0250e+01
    """
    with open(file_tim, 'w') as f:
        f.write(str_tim)
    
    data_tim = hcdfm.TimModel(file_tim)
    
    refdate = '2016-01-01'
    tim_pd = dfmt.TimModel_to_DataFrame(data_tim, parse_column_labels=True, refdate=refdate)
    
    assert tim_pd.index[0] == pd.Timestamp(refdate)
    assert len(tim_pd) == 8
    assert tim_pd.columns[-1] == 'Temperature (deg)'
    
    data_tim_out = dfmt.DataFrame_to_TimModel(tim_pd, refdate=refdate)
    assert len(data_tim_out.comments) == 4
    assert len(data_tim_out.timeseries) == 8



import xarray as xr
import xugrid as xu
import pandas as pd
import logging

__all__ = ["get_ncvarproperties",
           "rename_waqvars",
           "rename_fouvars",
           ]

logger = logging.getLogger(__name__)


def get_ncvarproperties(data_xr):
    if not isinstance(data_xr,(xr.Dataset,xu.UgridDataset)):
        raise TypeError('data_xr should be of type xr.Dataset or xu.UgridDataset')
    
    nc_varkeys = data_xr.variables.mapping.keys()
    
    list_varattrs_pd = []
    for varkey in nc_varkeys:
        varattrs_pd = pd.DataFrame({varkey:data_xr.variables.mapping[varkey].attrs}).T
        varattrs_pd[['shape','dimensions','dtype']] = 3*[''] #set dtype as str (float will raise an error when putting tuple in there)
        varattrs_pd.at[varkey,'shape'] = data_xr[varkey].shape
        varattrs_pd.at[varkey,'dimensions'] = data_xr.variables[varkey].dims
        varattrs_pd.at[varkey,'dtype'] = data_xr.variables[varkey].dtype
        list_varattrs_pd.append(varattrs_pd)
    
    vars_pd = pd.concat(list_varattrs_pd,axis=0)
    vars_pd[vars_pd.isnull()] = '' #avoid nan values
    
    data_xr.close()

    return vars_pd


def rename_waqvars(ds:(xr.Dataset,xu.UgridDataset)):
    """
    Rename all water quality variables in a dataset (like mesh2d_water_quality_output_24) to their long_name attribute (like mesh2d_DOscore)
    
    Parameters
    ----------
    ds : (xr.Dataset,xu.UgridDataset)
        DESCRIPTION.

    Returns
    -------
    ds : TYPE
        DESCRIPTION.

    """
    #TODO: results also in variable "mesh2d_Water quality mass balance areas" (with spaces), report in FM issue (remove spaces from long_name attr)
    
    if hasattr(ds,'grid'): #append gridname (e.g. mesh2d) in case of mapfile
        varn_prepend = f'{ds.grid.name}_'
    else:
        varn_prepend = ''
    list_waqvars = [i for i in ds.data_vars if 'water_quality_' in i] #water_quality_output and water_quality_stat
    rename_dict = {waqvar:varn_prepend+ds[waqvar].attrs['long_name'] for waqvar in list_waqvars}
    
    if len(rename_dict) == 0: #early return to silence "FutureWarning: The default dtype for empty Series will be 'object' instead of 'float64' in a future version. Specify a dtype explicitly to silence this warning."
        return ds
    
    #prevent renaming duplicate long_names
    rename_pd = pd.Series(rename_dict)
    if rename_pd.duplicated().sum():
        duplicated_pd = rename_pd.loc[rename_pd.duplicated(keep=False)]
        print(UserWarning(f'duplicate long_name attributes found with dfmt.rename_waqvars(), renaming only first variable:\n{duplicated_pd}'))
        rename_dict = rename_pd.loc[~rename_pd.duplicated()].to_dict()
    
    ds = ds.rename(rename_dict)
    return ds


def rename_fouvars(ds:(xr.Dataset,xu.UgridDataset), drop_tidal_times:bool = True):
    """
    Rename all fourier variables in a dataset (like mesh2d_fourier033_amp) to a unique name containing gridname/quantity/analysistype/tstart/tstop
    
    Parameters
    ----------
    ds : (xr.Dataset,xu.UgridDataset)
        DESCRIPTION.

    Returns
    -------
    ds : TYPE
        DESCRIPTION.

    """
    
    file_freqs = 'https://raw.githubusercontent.com/Deltares/hatyan/main/hatyan/data/data_foreman_frequencies.txt' #TODO: fix hatyan dependency (MSQM and M1 were also added, but file is not used by hatyan, so might disappear one day)
    freqs_pd = pd.read_csv(file_freqs, names=['freq','dependents'], sep="\\s+", comment='#')
    freqs_pd['angfreq'] = freqs_pd['freq'] * 360 #deg/hr
    
    gridname = ds.grid.name
    list_fouvars = [i for i in ds.data_vars if '_fourier' in i] #water_quality_output and water_quality_stat
    
    rename_dict = {}
    for fouvar in list_fouvars:
        fouvar_attrs_lower = {k.lower():v for k,v in ds[fouvar].attrs.items()}
        fouvar_lowerattrs = ds[fouvar].assign_attrs(fouvar_attrs_lower) #to avoid case issues
        
        #quantity
        long_name = fouvar_lowerattrs.attrs['long_name']
        long_name_noprefix = long_name.split(': ')[1]
        quantity_long = long_name_noprefix.split(',')[0]
        quantity_dict = {'water level':'s1', #dict based on https://svn.oss.deltares.nl/repos/delft3d/trunk/src/engines_gpl/dflowfm/packages/dflowfm_kernel/src/dflowfm_kernel/prepost/fourier_analysis.f90
                         #'energy head':'s1', #TODO: duplicate namfun/dictvalue is not convenient
                         'wind speed':'ws',
                         'U-component of cell-centre velocity':'ux',
                         'V-component of cell-centre velocity':'uy',
                         'U-component velocity, column average':'uxa',
                         'V-component velocity, column average':'uya',
                         'velocity magnitude':'uc',
                         #'':'r1', #TODO: unclear which namfun/dictvalue corresponds (trim(namcon(gdfourier%fconno(ifou))))
                         'velocity':'u1',
                         'unit discharge':'qx',
                         'bed stress':'ta',
                         'freeboard':'fb',
                         'waterdepth_on_ground':'wdog',
                         'volume_on_ground':'vog',
                         'discharge through flow link':'q1',
                         'water level at flow link':'su1',
                         'temperature':'tem', #not clear from fourier_analysis.f90, ct in user manual C.13
                         'salt':'sal', #not clear from fourier_analysis.f90, cs in user manual C.13
                         }
        if quantity_long not in quantity_dict.keys():
            raise KeyError(f'quantity_dict does not yet contain quantity for: {quantity_long}')
        quantity = quantity_dict[quantity_long]
        
        #analysistype
        istidal = False
        if hasattr(fouvar_lowerattrs,'frequency_degrees_per_hour'):
            if fouvar_lowerattrs.attrs['frequency_degrees_per_hour'] > 0: #wl mean with numcyc=0 has frequency attribute (wl min with numcyc=0 does not)
                istidal = True #for tidal components with frequency >0
        
        if istidal: #for tidal analysistype
            tidepart = fouvar.split('_')[-1] # amp/phs
            freq = fouvar_lowerattrs.attrs['frequency_degrees_per_hour']
            compidx_closestfreq = (freqs_pd['angfreq'] - freq).abs().argmin()
            compname = freqs_pd.index[compidx_closestfreq] #M2/NU2
            analysistype = tidepart+compname
            logger.warning('tidal components found in foufile, matching frequency with online list to get component names, '
                           'which might go wrong. Also, be aware that v0 and knfac columns from fourier inputfile are '
                           'not available in fourier output, so it is not clear whether to correct for these.')
        else: #for all other quantities
            fouvar_splitted = fouvar.split('_')
            analysistype = ''.join(fouvar_splitted[2:]) #min/max/mean. min_depth/max_depth etc are converted to mindepth/maxdepth
            
        #tstart/tstop
        refdate = pd.Timestamp(str(fouvar_lowerattrs.attrs['reference_date_in_yyyymmdd']))
        if hasattr(fouvar_lowerattrs,'starttime_fourier_analysis_in_minutes_since_reference_date'):
            tstart_min = fouvar_lowerattrs.attrs['starttime_fourier_analysis_in_minutes_since_reference_date']
            tstop_min = fouvar_lowerattrs.attrs['stoptime_fourier_analysis_in_minutes_since_reference_date']
        elif hasattr(fouvar_lowerattrs,'starttime_min_max_analysis_in_minutes_since_reference_date'):
            tstart_min = fouvar_lowerattrs.attrs['starttime_min_max_analysis_in_minutes_since_reference_date']
            tstop_min = fouvar_lowerattrs.attrs['stoptime_min_max_analysis_in_minutes_since_reference_date']
        else:
            raise AttributeError(f'starttime/stoptime attribute not found in fouvar:\n{fouvar_lowerattrs.attrs}')
        tstart_str = (refdate + pd.Timedelta(minutes=tstart_min)).strftime('%Y%m%d%H%M%S')
        tstop_str = (refdate + pd.Timedelta(minutes=tstop_min)).strftime('%Y%m%d%H%M%S')
        
        rename_dict[fouvar] = f'{gridname}_{quantity}_{analysistype}_{tstart_str}_{tstop_str}'
        if istidal and drop_tidal_times:
            rename_dict[fouvar] = f'{gridname}_{quantity}_{analysistype}' #TODO: might cause conflicting variable names if one component is analysed for multiple periods or if component is not defined in frequency list. Add duplicate check like rename_waqvars() that provides some basic info for debugging.
    
    ds = ds.rename(rename_dict)
    return ds




# -*- coding: utf-8 -*-
"""
Created on Thu Jun 22 09:41:49 2023

@author: veenstra
"""

import os
import pytest
import pandas as pd
import cdsapi
from dfm_tools.download import (cds_credentials,
                                cds_set_credentials,
                                cds_set_credentials_rcfile,
                                cds_remove_credentials_raise,
                                copernicusmarine_credentials,
                                copernicusmarine_get_dataset_id,
                                )
import dfm_tools as dfmt
import xarray as xr
import glob
import numpy as np
from unittest.mock import patch


def get_cds_url_key():
    try:
        cds_url, cds_apikey, _ = cdsapi.api.get_url_key_verify(url=None, key=None, verify=None)
    except Exception as e:
        if "Missing/incomplete configuration file" in str(e):
            cds_url = None
            cds_apikey = None
        else:
            raise e
    
    return cds_url, cds_apikey


def set_cds_credentials_ifnot_none(cds_url, cds_apikey):
    if None not in [cds_url, cds_apikey]:
        cds_set_credentials(cds_url, cds_apikey)
   

@pytest.mark.requiressecrets
@pytest.mark.unittest
def test_cds_credentials():
    cds_credentials()


@patch("getpass.getpass")
def test_cds_credentials_prompt(getpass):
    # backup credentials and remove credentials envvars and file
    cds_url, cds_apikey = get_cds_url_key()
    with pytest.raises(ValueError):
        cds_remove_credentials_raise()
    
    # provide apikey to cds_credentials() prompt
    getpass.return_value = cds_apikey
    cds_credentials()

    # restore credentials file/envvars
    set_cds_credentials_ifnot_none(cds_url, cds_apikey)


@pytest.mark.requiressecrets
@pytest.mark.unittest
def test_cds_credentials_onlykey_envvars():
    # backup credentials and remove credentials envvars and file
    cds_url, cds_apikey = get_cds_url_key()
    with pytest.raises(ValueError):
        cds_remove_credentials_raise()
    
    # test
    assert "CDSAPI_URL" not in os.environ.keys()
    os.environ["CDSAPI_KEY"] = cds_apikey
    cds_credentials()
    
    # restore credentials file/envvars
    set_cds_credentials_ifnot_none(cds_url, cds_apikey)


@pytest.mark.unittest
def test_cds_credentials_newurl_incorrectkey_rcfile():
    # backup credentials and remove credentials envvars and file
    cds_url, cds_apikey = get_cds_url_key()
    with pytest.raises(ValueError):
        cds_remove_credentials_raise()
    
    # test
    cds_url_temp = "https://cds.climate.copernicus.eu/api"
    cds_apikey_temp = "INCORRECT-APIKEY"
    cds_set_credentials_rcfile(cds_url_temp, cds_apikey_temp)
    with pytest.raises(ValueError) as e:
        cds_credentials()
    assert "Authentication failed" in str(e.value)
    assert "The CDS/ECMWF apikey file (~/.cdsapirc) was deleted" in str(e.value)    
    
    # restore credentials file/envvars
    set_cds_credentials_ifnot_none(cds_url, cds_apikey)


@pytest.mark.unittest
def test_cds_credentials_newurl_incorrectkey_envvars():
    # backup credentials and remove credentials envvars and file
    cds_url, cds_apikey = get_cds_url_key()
    with pytest.raises(ValueError):
        cds_remove_credentials_raise()
    
    # test
    os.environ["CDSAPI_URL"] = "https://cds.climate.copernicus.eu/api"
    os.environ["CDSAPI_KEY"] = "INCORRECT-APIKEY"
    with pytest.raises(ValueError) as e:
        cds_credentials()
    assert "Authentication failed" in str(e.value)
    assert "The CDS/ECMWF apikey file (~/.cdsapirc) was deleted" in str(e.value)
    
    # restore credentials file/envvars
    set_cds_credentials_ifnot_none(cds_url, cds_apikey)


@pytest.mark.unittest
def test_cds_credentials_oldurl_incorrectkey_rcfile():
    # backup credentials and remove credentials envvars and file
    cds_url, cds_apikey = get_cds_url_key()
    with pytest.raises(ValueError):
        cds_remove_credentials_raise()
    
    # test
    cds_url_temp = "https://cds.climate.copernicus.eu/api/v2"
    cds_apikey_temp = "INCORRECT-APIKEY"
    cds_set_credentials_rcfile(cds_url_temp, cds_apikey_temp)
    with pytest.raises(ValueError) as e:
        with pytest.warns(UserWarning) as w:
            cds_credentials()
    assert "This means an old/incorrect CDSAPI URL was found" in str(e.value)
    assert "The CDS/ECMWF apikey file (~/.cdsapirc) was deleted" in str(e.value)
    assert "404 Client Error: Not Found for url:" in str(w[0].message)
    
    # test
    cds_url_temp = "https://cds-beta.climate.copernicus.eu/api"
    cds_apikey_temp = "INCORRECT-APIKEY"
    cds_set_credentials_rcfile(cds_url_temp, cds_apikey_temp)
    with pytest.raises(ValueError) as e:
        with pytest.warns(UserWarning) as w:
            cds_credentials()
    assert "This means an old CDSAPI URL was found" in str(e.value)
    assert "The CDS/ECMWF apikey file (~/.cdsapirc) was deleted" in str(e.value)
    assert "certificate verify failed" in str(w[0].message)
    
    # restore credentials file/envvars
    set_cds_credentials_ifnot_none(cds_url, cds_apikey)


@pytest.mark.unittest
def test_cds_credentials_oldurl_incorrectkey_envvars():
    # backup credentials and remove credentials envvars and file
    cds_url, cds_apikey = get_cds_url_key()
    with pytest.raises(ValueError):
        cds_remove_credentials_raise()
    
    # test
    os.environ["CDSAPI_URL"] = "https://cds.climate.copernicus.eu/api/v2"
    os.environ["CDSAPI_KEY"] = "INCORRECT-APIKEY"
    with pytest.raises(ValueError) as e:
        with pytest.warns(UserWarning) as w:
            cds_credentials()
    assert "This means an old/incorrect CDSAPI URL was found" in str(e.value)
    assert "The CDS/ECMWF apikey file (~/.cdsapirc) was deleted" in str(e.value)
    assert "404 Client Error: Not Found for url:" in str(w[0].message)
    
    # test
    os.environ["CDSAPI_URL"] = "https://cds-beta.climate.copernicus.eu/api"
    os.environ["CDSAPI_KEY"] = "INCORRECT-APIKEY"
    with pytest.raises(ValueError) as e:
        with pytest.warns(UserWarning) as w:
            cds_credentials()
    assert "This means an old CDSAPI URL was found" in str(e.value)
    assert "The CDS/ECMWF apikey file (~/.cdsapirc) was deleted" in str(e.value)
    assert "certificate verify failed" in str(w[0].message)

    # restore credentials file/envvars
    set_cds_credentials_ifnot_none(cds_url, cds_apikey)


@pytest.mark.unittest
def test_cds_credentials_newurl_oldkey_rcfile():
    # backup credentials and remove credentials envvars and file
    cds_url, cds_apikey = get_cds_url_key()
    with pytest.raises(ValueError):
        cds_remove_credentials_raise()
    
    # test
    cds_url_temp = "https://cds.climate.copernicus.eu/api"
    cds_apikey_temp = "olduid:old-api-key"
    cds_set_credentials_rcfile(cds_url_temp, cds_apikey_temp)
    with pytest.raises(ValueError) as e:
        cds_credentials()
    assert "This means an old CDSAPI KEY was found (with :)" in str(e.value)
    assert "The CDS/ECMWF apikey file (~/.cdsapirc) was deleted" in str(e.value)
    
    # restore credentials file/envvars
    set_cds_credentials_ifnot_none(cds_url, cds_apikey)


@pytest.mark.unittest
def test_cds_credentials_newurl_oldkey_envvars():
    # backup credentials and remove credentials envvars and file
    cds_url, cds_apikey = get_cds_url_key()
    with pytest.raises(ValueError):
        cds_remove_credentials_raise()
    
    # test
    os.environ["CDSAPI_URL"] = "https://cds.climate.copernicus.eu/api"
    os.environ["CDSAPI_KEY"] = "olduid:old-api-key"
    with pytest.raises(ValueError) as e:
        cds_credentials()
    assert "This means an old CDSAPI KEY was found (with :)" in str(e.value)
    assert "The CDS/ECMWF apikey file (~/.cdsapirc) was deleted" in str(e.value)
    
    # restore credentials file/envvars
    set_cds_credentials_ifnot_none(cds_url, cds_apikey)


@pytest.mark.requiressecrets
@pytest.mark.unittest
def test_copernicusmarine_credentials():
    copernicusmarine_credentials()


@pytest.mark.requiressecrets
@pytest.mark.unittest
def test_copernicusmarine_get_dataset_id():
    # with string datetimes, these are parsed to pandas timestamps in
    # copernicusmarine_get_product()
    date_min = '2010-01-01'
    date_max = '2010-01-02'
    date_args = dict(date_min=date_min, date_max=date_max)
    dataset_id = copernicusmarine_get_dataset_id(varkey='bottomT', **date_args)
    assert dataset_id == 'cmems_mod_glo_phy_my_0.083deg_P1D-m'
    dataset_id = copernicusmarine_get_dataset_id(varkey='no3', **date_args)
    assert dataset_id == 'cmems_mod_glo_bgc_my_0.25deg_P1D-m'
    dataset_id = copernicusmarine_get_dataset_id(varkey='phyc', **date_args)
    assert dataset_id == 'cmems_mod_glo_bgc_my_0.25deg_P1M-m'
    
    # with pandas timestamps
    date_min = pd.Timestamp.today()
    date_max = pd.Timestamp.today() + pd.Timedelta(days=1)
    date_args = dict(date_min=date_min, date_max=date_max)
    dataset_id = copernicusmarine_get_dataset_id(varkey='tob', **date_args)
    assert dataset_id == 'cmems_mod_glo_phy_anfc_0.083deg_P1D-m'
    dataset_id = copernicusmarine_get_dataset_id(varkey='no3', **date_args)
    assert dataset_id == 'cmems_mod_glo_bgc-nut_anfc_0.25deg_P1D-m'


@pytest.mark.parametrize("varkey", [pytest.param(x, id=x) for x in ['bottomT','no3','so','tob']])
@pytest.mark.requiressecrets
@pytest.mark.unittest
def test_download_cmems(tmp_path, varkey):
    """
    Test whether downloading cmems data works properly.
    Especially whether time/spatial extents are correct (buffered)
    and whether daily means are corrected with a 12 hour offset.
    
    the variables retrieved are arbitrary, but are present in the respective datasets
    avaliable variables differ per product, examples are ['bottomT','mlotst','siconc','sithick','so','thetao','uo','vo','usi','vsi','zos','no3'].
    More info on https://data.marine.copernicus.eu/products
    
    In the variable/dataset selection it is ensured to cover both
    reanalysis and analysisforecasts, both daily and monthly means,
    both bio and phy, both 0.25deg and 0.083deg spatial resolutions
    and both variables including and excluding depth
    """
    
    # deliberately take inconvenient time/spatial subset to test if
    # coordinates_selection_method='outside' works properly
    longitude_min, longitude_max, latitude_min, latitude_max =    2.001,   3.001,  51.001, 52.001 #test domain
    dataset_id_dict = {'bottomT':'cmems_mod_glo_phy_my_0.083deg_P1D-m', # phy my daily mean no_depth
                       'no3':'cmems_mod_glo_bgc_my_0.25deg_P1D-m', # bio my daily mean wi_depth
                       'so':'cmems_mod_glo_phy_my_0.083deg_P1M-m', # phy my monthly mean wi_depth
                       'tob':'cmems_mod_glo_phy_anfc_0.083deg_P1D-m', # phy anfc daily mean no_depth
                       }
    file_prefix = 'cmems_'
    dataset_id = dataset_id_dict[varkey]
    # freq is converted from "D" to "M" automatically in for monthly datasets
    freq = "D"

    if "_my_" in dataset_id:
        date_min = '2010-01-01 01:00'
        date_max = '2010-01-01 23:00'
    elif "_anfc_"in dataset_id:
        date_today = pd.Timestamp.today().floor("1D")
        date_min = date_today + pd.Timedelta(hours=1)
        date_max = date_today + pd.Timedelta(hours=23)
    
    if varkey == 'so':
        # monthly mean, times are at midnight (no offset applied)
        times_expected = ['2010-01-01 00:00:00', '2010-02-01 00:00:00']
        fnames_expected = ['cmems_so_2010-01.nc', 'cmems_so_2010-02.nc']
    elif varkey == 'no3':
        # daily mean, so offset of 12 hours is applied
        times_expected = ['2009-12-31 12:00:00',
                          '2010-01-01 12:00:00',
                          '2010-01-02 12:00:00',
                          ]
        fnames_expected = ['cmems_no3_2009-12-31.nc',
                           'cmems_no3_2010-01-01.nc',
                           'cmems_no3_2010-01-02.nc',
                           ]
    elif varkey == 'bottomT':
        # daily mean, so offset of 12 hours is applied
        times_expected = ['2009-12-31 12:00:00',
                          '2010-01-01 12:00:00',
                          '2010-01-02 12:00:00']
        fnames_expected = ['cmems_bottomT_2009-12-31.nc',
                           'cmems_bottomT_2010-01-01.nc',
                           'cmems_bottomT_2010-01-02.nc',
                           ]
    elif varkey == 'tob':
        # daily mean, so offset of 12 hours is applied
        datetime_first = date_today - pd.Timedelta(hours=12)
        date_first = datetime_first.date()
        times_expected = [str(datetime_first + x*pd.Timedelta("1D")) for x in [0,1,2]]
        fnames_dates = [str(date_first + x*pd.Timedelta("1D")) for x in [0,1,2]]
        fnames_expected = [f"cmems_tob_{x}.nc" for x in fnames_dates]
    
    if '0.25deg' in dataset_id:
        lon_max_exp = 3.25
        lat_max_exp = 52.25
    elif '0.083deg' in dataset_id:
        lon_max_exp = 3.08333
        lat_max_exp = 52.083332
        
    dfmt.download_CMEMS(varkey=varkey,
                        longitude_min=longitude_min, longitude_max=longitude_max, latitude_min=latitude_min, latitude_max=latitude_max,
                        date_min=date_min, date_max=date_max, freq=freq,
                        # speed up tests by supplying datset_id
                        dataset_id=dataset_id,
                        dir_output=tmp_path, file_prefix=file_prefix, overwrite=True)

    # open downloaded files
    file_nc_pat = os.path.join(tmp_path, f"{file_prefix}{varkey}*.nc")
    fname_list = [os.path.basename(x) for x in glob.glob(file_nc_pat)]
    fname_list.sort()
    ds = xr.open_mfdataset(file_nc_pat)
    times_actual = ds.time.to_pandas().dt.strftime("%Y-%m-%d %H:%M:%S").tolist()
    
    # assertions
    assert varkey in set(ds.variables)
    assert ds.sizes["time"] == len(times_expected)
    assert fname_list == fnames_expected
    assert times_actual == times_expected
    assert np.isclose(ds.longitude.to_numpy().min(), 2)
    assert np.isclose(ds.longitude.to_numpy().max(), lon_max_exp)
    assert np.isclose(ds.latitude.to_numpy().min(), 51)
    assert np.isclose(ds.latitude.to_numpy().max(), lat_max_exp)


@pytest.mark.unittest
def test_download_hycom(tmp_path):
    # domain
    longitude_min, longitude_max, latitude_min, latitude_max =    2,   3,  51, 52 #test domain
    date_min = '2010-01-01'
    date_max = '2010-01-02'
    varlist_hycom = ['surf_el']#'water_temp'] #['tau','water_u','water_v','water_temp','salinity','surf_el']
    
    for varkey in varlist_hycom:
        # Path(dir_output).mkdir(parents=True, exist_ok=True)
        period_range_years = pd.period_range(date_min,date_max,freq='Y')
        dataset_url = [f'https://tds.hycom.org/thredds/dodsC/GLBu0.08/expt_19.1/{year}' for year in period_range_years] #list is possible with hycom, since it uses xr.open_mfdataset()
        # temporary fix to avoid RuntimeError: NetCDF: file not found
        # https://github.com/Deltares/dfm_tools/issues/1048
        file_prefix = 'hycom_'
        dfmt.download_OPeNDAP(dataset_url=dataset_url,
                              varkey=varkey,
                              longitude_min=longitude_min, longitude_max=longitude_max, latitude_min=latitude_min, latitude_max=latitude_max,
                              date_min=date_min, date_max=date_max,
                              dir_output=tmp_path, file_prefix=file_prefix, overwrite=True)

    # assert downloaded files
    file_nc_pat = os.path.join(tmp_path, "*.nc")
    ds = xr.open_mfdataset(file_nc_pat)
    assert ds.sizes["time"] == 2
    assert ds.time.to_pandas().iloc[0] == pd.Timestamp('2010-01-01')
    assert ds.time.to_pandas().iloc[-1] == pd.Timestamp('2010-01-02')


@pytest.mark.requiressecrets
@pytest.mark.unittest
def test_download_era5_unsupported_varkey():
    date_min = '2010-01-31'
    date_max = '2010-02-01'
    longitude_min, longitude_max, latitude_min, latitude_max = 2, 3, 51, 52
    varkey = 'unexisting'
    with pytest.raises(KeyError) as e:
        dfmt.download_ERA5(
            varkey, 
            longitude_min=longitude_min, longitude_max=longitude_max,
            latitude_min=latitude_min, latitude_max=latitude_max,
            date_min=date_min, date_max=date_max,
            dir_output='.', overwrite=True)
    assert '"unexisting" not available' in str(e.value)


@pytest.mark.requiressecrets
@pytest.mark.unittest
def test_download_era5_incorrect_times():
    date_min = '2010-01-01'
    date_max = '2009-01-01'
    longitude_min, longitude_max, latitude_min, latitude_max = 2, 3, 51, 52
    varkey = 'msl'
    with pytest.raises(ValueError) as e:
        dfmt.download_ERA5(
            varkey, 
            longitude_min=longitude_min, longitude_max=longitude_max,
            latitude_min=latitude_min, latitude_max=latitude_max,
            date_min=date_min, date_max=date_max,
            dir_output='.', overwrite=True)
    assert 'resulted in empty period_range' in str(e.value)


@pytest.mark.requiressecrets
@pytest.mark.unittest
@pytest.mark.era5slow # temporarily skip these on github
@pytest.mark.timeout(60) # useful since CDS downloads are terribly slow sometimes, so skip in that case
def test_download_era5(file_nc_era5_pattern):
    # file_nc_era5_pattern comes from conftest.py
    file_list = glob.glob(file_nc_era5_pattern)
    assert len(file_list) == 2
    
    ds = xr.open_mfdataset(file_nc_era5_pattern)
    
    # datasets retrieved with intermediate CDS had expver dimension causing issues
    # if this fails, update the docstring of preprocess_ERA5
    assert 'expver' not in ds.dims
    # if this fails, update preprocess_ERA5 to not add coordinate varialbe
    assert 'expver' in ds.coords
    
    # datasets retrieved with new CDS have valid_time instead of time dim/var
    assert 'valid_time' in ds.dims # TODO: if this fails, remove the exception below and in preprocess_ERA5
    timedim = 'valid_time'
    
    assert ds.sizes[timedim] == 1416
    assert ds[timedim].to_pandas().iloc[0] == pd.Timestamp('2010-01-01')
    assert ds[timedim].to_pandas().iloc[-1] == pd.Timestamp('2010-02-28 23:00')
    
    # datasets retrieved with new CDS are float32 instead of scaled ints
    # ints raised problem in https://github.com/Deltares/dfm_tools/issues/239
    msl_encoding = ds['msl'].encoding
    assert str(msl_encoding['dtype']) == 'float32'
    assert 'scale_factor' not in msl_encoding.keys()



# -*- coding: utf-8 -*-
"""
Created on Sun Jul  9 22:17:16 2023

@author: veenstra
"""

import pytest
import xarray as xr
import xugrid as xu
import dfm_tools as dfmt


@pytest.mark.unittest
def test_data_map():
    func_list = [dfmt.data.fm_grevelingen_map, 
                 dfmt.data.fm_grevelingen_net, 
                 dfmt.data.fm_curvedbend_map, 
                 dfmt.data.fm_westernscheldt_map, 
                 dfmt.data.fm_westernscheldt_fou, 
                 dfmt.data.fm_westernscheldt_rst, 
                 dfmt.data.d3d_westernscheldt_trim,
                 dfmt.data.d3d_curvedbend_trim]
    
    for func in func_list:
        file_nc = func(return_filepath=True)
        uds = func()
        assert isinstance(file_nc,str)
        assert isinstance(uds,xu.UgridDataset)


@pytest.mark.unittest
def test_data_his():
    func_list = [dfmt.data.fm_grevelingen_his, 
                 dfmt.data.fm_curvedbend_his,
                 dfmt.data.fm_westernscheldt_his,
                 dfmt.data.d3d_curvedbend_trih]
    
    for func in func_list:
        file_nc = func(return_filepath=True)
        uds = func()
        assert isinstance(file_nc,str)
        assert isinstance(uds,xr.Dataset)


@pytest.mark.unittest
def test_gshhs_coastlines_shp():
    gshhs_dir = dfmt.data.gshhs_coastlines_shp()
    assert isinstance(gshhs_dir,str)
